from collections.abc import Callable
from fractions import Fraction
from typing import Any, LiteralString, assert_type, type_check_only

import numpy as np
import numpy.typing as npt

f8: np.float64
AR_LIKE_b: list[bool]
AR_LIKE_i8: list[int]
AR_LIKE_f8: list[float]
AR_LIKE_c16: list[complex]
AR_LIKE_O: list[Fraction]

AR_u1: npt.NDArray[np.uint8]
AR_i8: npt.NDArray[np.int64]
AR_f2: npt.NDArray[np.float16]
AR_f4: npt.NDArray[np.float32]
AR_f8: npt.NDArray[np.float64]
AR_f10: npt.NDArray[np.longdouble]
AR_c8: npt.NDArray[np.complex64]
AR_c16: npt.NDArray[np.complex128]
AR_c20: npt.NDArray[np.clongdouble]
AR_m: npt.NDArray[np.timedelta64]
AR_M: npt.NDArray[np.datetime64]
AR_O: npt.NDArray[np.object_]
AR_b: npt.NDArray[np.bool]
AR_U: npt.NDArray[np.str_]
CHAR_AR_U: np.char.chararray[tuple[Any, ...], np.dtype[np.str_]]  # type: ignore[deprecated]

AR_f8_1d: np.ndarray[tuple[int], np.dtype[np.float64]]
AR_f8_2d: np.ndarray[tuple[int, int], np.dtype[np.float64]]
AR_f8_3d: np.ndarray[tuple[int, int, int], np.dtype[np.float64]]
AR_c16_1d: np.ndarray[tuple[int], np.dtype[np.complex128]]

AR_b_list: list[npt.NDArray[np.bool]]

@type_check_only
def func(a: np.ndarray, posarg: bool = ..., /, arg: int = ..., *, kwarg: str = ...) -> np.ndarray: ...
@type_check_only
def func_f8(a: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]: ...

###

# vectorize
vectorized_func: np.vectorize
assert_type(vectorized_func.pyfunc, Callable[..., Any])
assert_type(vectorized_func.cache, bool)
assert_type(vectorized_func.signature, LiteralString | None)
assert_type(vectorized_func.otypes, LiteralString | None)
assert_type(vectorized_func.excluded, set[int | str])
assert_type(vectorized_func.__doc__, str | None)
assert_type(vectorized_func([1]), Any)
assert_type(np.vectorize(int), np.vectorize)
assert_type(
    np.vectorize(int, otypes="i", doc="doc", excluded=(), cache=True, signature=None),
    np.vectorize,
)

# rot90
assert_type(np.rot90(AR_f8_1d), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.rot90(AR_f8, k=2), npt.NDArray[np.float64])
assert_type(np.rot90(AR_LIKE_f8, axes=(0, 1)), np.ndarray)

# flip
assert_type(np.flip(AR_f8_1d), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.flip(AR_f8, axis=(0, 1)), npt.NDArray[np.float64])
assert_type(np.flip(AR_LIKE_f8, axis=0), np.ndarray)

# iterable
assert_type(np.iterable(1), bool)
assert_type(np.iterable([1]), bool)

# average
assert_type(np.average(AR_f8_2d), np.float64)
assert_type(np.average(AR_f8_2d, axis=1), npt.NDArray[np.float64])
assert_type(np.average(AR_f8_2d, keepdims=True), np.ndarray[tuple[int, int], np.dtype[np.float64]])
assert_type(np.average(AR_f8), np.float64)
assert_type(np.average(AR_f8, axis=1), npt.NDArray[np.float64])
assert_type(np.average(AR_f8, keepdims=True), npt.NDArray[np.float64])
assert_type(np.average(AR_f8, returned=True), tuple[np.float64, np.float64])
assert_type(np.average(AR_f8, axis=1, returned=True), tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]])
assert_type(np.average(AR_f8, keepdims=True, returned=True), tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]])
assert_type(np.average(AR_LIKE_f8), np.float64)
assert_type(np.average(AR_LIKE_f8, weights=AR_f8), np.float64)
assert_type(np.average(AR_LIKE_f8, axis=1), npt.NDArray[np.float64])
assert_type(np.average(AR_LIKE_f8, keepdims=True), npt.NDArray[np.float64])
assert_type(np.average(AR_LIKE_f8, returned=True), tuple[np.float64, np.float64])
assert_type(np.average(AR_LIKE_f8, axis=1, returned=True), tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]])
assert_type(np.average(AR_LIKE_f8, keepdims=True, returned=True), tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]])
assert_type(np.average(AR_O), Any)
assert_type(np.average(AR_O, axis=1), np.ndarray)
assert_type(np.average(AR_O, keepdims=True), np.ndarray)
assert_type(np.average(AR_O, returned=True), tuple[Any, Any])
assert_type(np.average(AR_O, axis=1, returned=True), tuple[np.ndarray, np.ndarray])
assert_type(np.average(AR_O, keepdims=True, returned=True), tuple[np.ndarray, np.ndarray])

# asarray_chkfinite
assert_type(np.asarray_chkfinite(AR_f8_1d), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.asarray_chkfinite(AR_f8), npt.NDArray[np.float64])
assert_type(np.asarray_chkfinite(AR_LIKE_f8), np.ndarray)
assert_type(np.asarray_chkfinite(AR_f8, dtype=np.float64), npt.NDArray[np.float64])
assert_type(np.asarray_chkfinite(AR_f8, dtype=float), np.ndarray)

# piecewise
assert_type(np.piecewise(AR_f8_1d, AR_b, [func]), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.piecewise(AR_f8, AR_b, [func]), npt.NDArray[np.float64])
assert_type(np.piecewise(AR_f8, AR_b, [func_f8]), npt.NDArray[np.float64])
assert_type(np.piecewise(AR_f8, AR_b_list, [func]), npt.NDArray[np.float64])
assert_type(np.piecewise(AR_f8, AR_b_list, [func_f8]), npt.NDArray[np.float64])
assert_type(np.piecewise(AR_f8, AR_b_list, [func], True, -1, kwarg=""), npt.NDArray[np.float64])
assert_type(np.piecewise(AR_f8, AR_b_list, [func], True, arg=-1, kwarg=""), npt.NDArray[np.float64])
assert_type(np.piecewise(AR_LIKE_f8, AR_b_list, [func]), np.ndarray)
assert_type(np.piecewise(AR_LIKE_f8, AR_b_list, [func_f8]), npt.NDArray[np.float64])

# extract
assert_type(np.extract(AR_i8, AR_f8), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.extract(AR_i8, AR_LIKE_b), np.ndarray[tuple[int], np.dtype[np.bool]])
assert_type(np.extract(AR_i8, AR_LIKE_i8), np.ndarray[tuple[int], np.dtype[np.int_]])
assert_type(np.extract(AR_i8, AR_LIKE_f8), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.extract(AR_i8, AR_LIKE_c16), np.ndarray[tuple[int], np.dtype[np.complex128]])

# select
assert_type(np.select([AR_b], [AR_f8_1d]), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.select([AR_b], [AR_f8]), npt.NDArray[np.float64])

# places
assert_type(np.place(AR_f8, mask=AR_i8, vals=5.0), None)

# copy
assert_type(np.copy(AR_LIKE_f8), np.ndarray)
assert_type(np.copy(AR_U), npt.NDArray[np.str_])
assert_type(np.copy(CHAR_AR_U, "K", subok=True), np.char.chararray[tuple[Any, ...], np.dtype[np.str_]])  # type: ignore[deprecated]
assert_type(np.copy(CHAR_AR_U, subok=True), np.char.chararray[tuple[Any, ...], np.dtype[np.str_]])  # type: ignore[deprecated]
# pyright correctly infers `NDArray[str_]` here
assert_type(np.copy(CHAR_AR_U), np.ndarray[Any, Any])  # pyright: ignore[reportAssertTypeFailure]

# gradient
assert_type(np.gradient(AR_f8_1d, 1), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(
    np.gradient(AR_f8_2d, [1, 2], [2, 3.5, 4]),
    tuple[
        np.ndarray[tuple[int, int], np.dtype[np.float64]],
        np.ndarray[tuple[int, int], np.dtype[np.float64]],
    ],
)
assert_type(
    np.gradient(AR_f8_3d),
    tuple[
        np.ndarray[tuple[int, int, int], np.dtype[np.float64]],
        np.ndarray[tuple[int, int, int], np.dtype[np.float64]],
        np.ndarray[tuple[int, int, int], np.dtype[np.float64]],
    ],
)
assert_type(np.gradient(AR_f8), np.ndarray[tuple[int], np.dtype[np.float64]] | Any)
assert_type(np.gradient(AR_LIKE_f8, edge_order=2), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.gradient(AR_LIKE_c16, axis=0), np.ndarray[tuple[int], np.dtype[np.complex128]])

# diff
assert_type(np.diff("git", n=0), str)
assert_type(np.diff(AR_f8), npt.NDArray[np.float64])
assert_type(np.diff(AR_f8_1d, axis=0), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.diff(AR_f8_2d, axis=0), np.ndarray[tuple[int, int], np.dtype[np.float64]])
assert_type(np.diff(AR_LIKE_f8, prepend=1.5), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.diff(AR_c16), npt.NDArray[np.complex128])
assert_type(np.diff(AR_c16_1d), np.ndarray[tuple[int], np.dtype[np.complex128]])
assert_type(np.diff(AR_LIKE_c16), np.ndarray[tuple[int], np.dtype[np.complex128]])

# interp
assert_type(np.interp(1, [1], AR_f8), np.float64)
assert_type(np.interp(1, [1], [1]), np.float64)
assert_type(np.interp(1, [1], AR_c16), np.complex128)
assert_type(np.interp(1, [1], [1j]), np.complex128)
assert_type(np.interp([1], [1], AR_f8), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.interp([1], [1], [1]),  np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.interp([1], [1], AR_c16), np.ndarray[tuple[int], np.dtype[np.complex128]])
assert_type(np.interp([1], [1], [1j]), np.ndarray[tuple[int], np.dtype[np.complex128]])

# angle
assert_type(np.angle(1), np.float64)
assert_type(np.angle(1, deg=True), np.float64)
assert_type(np.angle(1j), np.float64)
assert_type(np.angle(f8), np.float64)
assert_type(np.angle(AR_b), npt.NDArray[np.float64])
assert_type(np.angle(AR_u1), npt.NDArray[np.float64])
assert_type(np.angle(AR_i8), npt.NDArray[np.float64])
assert_type(np.angle(AR_f2), npt.NDArray[np.float16])
assert_type(np.angle(AR_f4), npt.NDArray[np.float32])
assert_type(np.angle(AR_c8), npt.NDArray[np.float32])
assert_type(np.angle(AR_f8), npt.NDArray[np.float64])
assert_type(np.angle(AR_c16), npt.NDArray[np.float64])
assert_type(np.angle(AR_f10), npt.NDArray[np.longdouble])
assert_type(np.angle(AR_c20), npt.NDArray[np.longdouble])
assert_type(np.angle(AR_f8_1d), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.angle(AR_c16_1d), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.angle(AR_LIKE_b), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.angle(AR_LIKE_i8), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.angle(AR_LIKE_f8), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.angle(AR_LIKE_c16), np.ndarray[tuple[int], np.dtype[np.float64]])

# unwrap
assert_type(np.unwrap(AR_f2), npt.NDArray[np.float16])
assert_type(np.unwrap(AR_f8), npt.NDArray[np.float64])
assert_type(np.unwrap(AR_f10), npt.NDArray[np.longdouble])
assert_type(np.unwrap(AR_O), npt.NDArray[np.object_])
assert_type(np.unwrap(AR_f8_1d), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.unwrap(AR_f8_2d), np.ndarray[tuple[int, int], np.dtype[np.float64]])
assert_type(np.unwrap(AR_f8_3d), np.ndarray[tuple[int, int, int], np.dtype[np.float64]])
assert_type(np.unwrap(AR_LIKE_b), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.unwrap(AR_LIKE_i8), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.unwrap(AR_LIKE_f8), np.ndarray[tuple[int], np.dtype[np.float64]])

# sort_complex
assert_type(np.sort_complex(AR_u1), npt.NDArray[np.complex64])
assert_type(np.sort_complex(AR_f8), npt.NDArray[np.complex128])
assert_type(np.sort_complex(AR_f10), npt.NDArray[np.clongdouble])
assert_type(np.sort_complex(AR_f8_1d), np.ndarray[tuple[int], np.dtype[np.complex128]])
assert_type(np.sort_complex(AR_c16_1d), np.ndarray[tuple[int], np.dtype[np.complex128]])

# trim_zeros
assert_type(np.trim_zeros(AR_f8), npt.NDArray[np.float64])
assert_type(np.trim_zeros(AR_LIKE_f8), list[float])

# cov
assert_type(np.cov(AR_f8_1d), np.ndarray[tuple[()], np.dtype[np.float64]])
assert_type(np.cov(AR_f8_2d), npt.NDArray[np.float64])
assert_type(np.cov(AR_f8), npt.NDArray[np.float64])
assert_type(np.cov(AR_f8, AR_f8), np.ndarray[tuple[int, int], np.dtype[np.float64]])
assert_type(np.cov(AR_c16, AR_c16), np.ndarray[tuple[int, int], np.dtype[np.complex128]])
assert_type(np.cov(AR_LIKE_f8), np.ndarray[tuple[()], np.dtype[np.float64]])
assert_type(np.cov(AR_LIKE_f8, AR_LIKE_f8), np.ndarray[tuple[int, int], np.dtype[np.float64]])
assert_type(np.cov(AR_LIKE_f8, dtype=np.float16), np.ndarray[tuple[()], np.dtype[np.float16]])
assert_type(np.cov(AR_LIKE_f8, AR_LIKE_f8, dtype=np.float32), np.ndarray[tuple[int, int], np.dtype[np.float32]])
assert_type(np.cov(AR_f8, AR_f8, dtype=float), np.ndarray[tuple[int, int]])
assert_type(np.cov(AR_LIKE_f8, dtype=float), np.ndarray[tuple[()]])
assert_type(np.cov(AR_LIKE_f8, AR_LIKE_f8, dtype=float), np.ndarray[tuple[int, int]])

# corrcoef
assert_type(np.corrcoef(AR_f8_1d), np.float64)
assert_type(np.corrcoef(AR_f8_2d), np.ndarray[tuple[int, int], np.dtype[np.float64]] | np.float64)
assert_type(np.corrcoef(AR_f8), np.ndarray[tuple[int, int], np.dtype[np.float64]] | np.float64)
assert_type(np.corrcoef(AR_f8, AR_f8), np.ndarray[tuple[int, int], np.dtype[np.float64]])
assert_type(np.corrcoef(AR_c16, AR_c16), np.ndarray[tuple[int, int], np.dtype[np.complex128]])
assert_type(np.corrcoef(AR_LIKE_f8), np.float64)
assert_type(np.corrcoef(AR_LIKE_f8, AR_LIKE_f8), np.ndarray[tuple[int, int], np.dtype[np.float64]])
assert_type(np.corrcoef(AR_LIKE_f8, dtype=np.float16), np.float16)
assert_type(np.corrcoef(AR_LIKE_f8, AR_LIKE_f8, dtype=np.float32), np.ndarray[tuple[int, int], np.dtype[np.float32]])
assert_type(np.corrcoef(AR_f8, AR_f8, dtype=float), np.ndarray[tuple[int, int]])
assert_type(np.corrcoef(AR_LIKE_f8, dtype=float), Any)
assert_type(np.corrcoef(AR_LIKE_f8, AR_LIKE_f8, dtype=float), np.ndarray[tuple[int, int]])

# window functions
assert_type(np.blackman(5), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.bartlett(6), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.hanning(4.5), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.hamming(0), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.kaiser(4, 5.9), np.ndarray[tuple[int], np.dtype[np.float64]])

# i0 (bessel function)
assert_type(np.i0(AR_i8), npt.NDArray[np.float64])

# sinc (cardinal sine function)
assert_type(np.sinc(1.0), np.float64)
assert_type(np.sinc(1j), np.complex128 | Any)
assert_type(np.sinc(AR_f8), npt.NDArray[np.float64])
assert_type(np.sinc(AR_c16), npt.NDArray[np.complex128])
assert_type(np.sinc(AR_LIKE_f8), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.sinc(AR_LIKE_c16), np.ndarray[tuple[int], np.dtype[np.complex128]])

# median
assert_type(np.median(AR_f8, keepdims=False), np.float64)
assert_type(np.median(AR_c16, overwrite_input=True), np.complex128)
# NOTE: Mypy incorrectly infers `Any`, but pyright behaves correctly.
assert_type(np.median(AR_m), np.timedelta64)  # type: ignore[assert-type]
assert_type(np.median(AR_O), Any)
assert_type(np.median(AR_f8, keepdims=True), npt.NDArray[np.float64])
assert_type(np.median(AR_f8, axis=0), npt.NDArray[np.float64])
assert_type(np.median(AR_c16, keepdims=True), npt.NDArray[np.complex128])
assert_type(np.median(AR_c16, axis=0), npt.NDArray[np.complex128])
assert_type(np.median(AR_LIKE_f8, keepdims=True), npt.NDArray[np.float64])
assert_type(np.median(AR_LIKE_c16, keepdims=True), npt.NDArray[np.complex128])
assert_type(np.median(AR_LIKE_f8, out=AR_c16), npt.NDArray[np.complex128])

# percentile
assert_type(np.percentile(AR_f8, 50), np.float64)
assert_type(np.percentile(AR_f8, 50, axis=1), npt.NDArray[np.float64])
assert_type(np.percentile(AR_f8, 50, axis=(1, 0)), npt.NDArray[np.float64])
assert_type(np.percentile(AR_f8, 50, keepdims=True), npt.NDArray[np.float64])
assert_type(np.percentile(AR_f8, 50, axis=0, keepdims=True), npt.NDArray[np.float64])
assert_type(np.percentile(AR_c16, 50), np.complex128)
assert_type(np.percentile(AR_m, 50), np.timedelta64)
assert_type(np.percentile(AR_M, 50, overwrite_input=True), np.datetime64)
assert_type(np.percentile(AR_O, 50), Any)
assert_type(np.percentile(AR_f8, [50]), npt.NDArray[np.float64])
assert_type(np.percentile(AR_f8, [50], axis=1), npt.NDArray[np.float64])
assert_type(np.percentile(AR_f8, [50], keepdims=True), npt.NDArray[np.float64])
assert_type(np.percentile(AR_c16, [50]), npt.NDArray[np.complex128])
assert_type(np.percentile(AR_m, [50]), npt.NDArray[np.timedelta64])
assert_type(np.percentile(AR_M, [50], method="nearest"), npt.NDArray[np.datetime64])
assert_type(np.percentile(AR_O, [50]), npt.NDArray[np.object_])
assert_type(np.percentile(AR_f8, [50], keepdims=True), npt.NDArray[np.float64])
assert_type(np.percentile(AR_f8, [50], out=AR_c16), npt.NDArray[np.complex128])

# quantile
assert_type(np.quantile(AR_f8, 0.50), np.float64)
assert_type(np.quantile(AR_f8, 0.50, axis=1), npt.NDArray[np.float64])
assert_type(np.quantile(AR_f8, 0.50, axis=(1, 0)), npt.NDArray[np.float64])
assert_type(np.quantile(AR_f8, 0.50, keepdims=True), npt.NDArray[np.float64])
assert_type(np.quantile(AR_f8, 0.50, axis=0, keepdims=True), npt.NDArray[np.float64])
assert_type(np.quantile(AR_c16, 0.50), np.complex128)
assert_type(np.quantile(AR_m, 0.50), np.timedelta64)
assert_type(np.quantile(AR_M, 0.50, overwrite_input=True), np.datetime64)
assert_type(np.quantile(AR_O, 0.50), Any)
assert_type(np.quantile(AR_f8, [0.50]), npt.NDArray[np.float64])
assert_type(np.quantile(AR_f8, [0.50], axis=1), npt.NDArray[np.float64])
assert_type(np.quantile(AR_f8, [0.50], keepdims=True), npt.NDArray[np.float64])
assert_type(np.quantile(AR_c16, [0.50]), npt.NDArray[np.complex128])
assert_type(np.quantile(AR_m, [0.50]), npt.NDArray[np.timedelta64])
assert_type(np.quantile(AR_M, [0.50], method="nearest"), npt.NDArray[np.datetime64])
assert_type(np.quantile(AR_O, [0.50]), npt.NDArray[np.object_])
assert_type(np.quantile(AR_f8, [0.50], keepdims=True), npt.NDArray[np.float64])
assert_type(np.quantile(AR_f8, [0.50], out=AR_c16), npt.NDArray[np.complex128])

# trapezoid
assert_type(np.trapezoid(AR_LIKE_f8), np.float64)
assert_type(np.trapezoid(AR_LIKE_f8, AR_LIKE_f8), np.float64)
assert_type(np.trapezoid(AR_LIKE_c16), np.complex128)
assert_type(np.trapezoid(AR_LIKE_c16, AR_LIKE_f8), np.complex128)
assert_type(np.trapezoid(AR_LIKE_f8, AR_LIKE_c16), np.complex128)
assert_type(np.trapezoid(AR_LIKE_O), float)
assert_type(np.trapezoid(AR_LIKE_O, AR_LIKE_f8), float)
assert_type(np.trapezoid(AR_f8), np.float64 | npt.NDArray[np.float64])
assert_type(np.trapezoid(AR_f8, AR_f8), np.float64 | npt.NDArray[np.float64])
assert_type(np.trapezoid(AR_c16), np.complex128 | npt.NDArray[np.complex128])
assert_type(np.trapezoid(AR_c16, AR_c16), np.complex128 | npt.NDArray[np.complex128])
# NOTE: Mypy incorrectly infers `Any`, but pyright behaves correctly.
assert_type(np.trapezoid(AR_m), np.timedelta64 | npt.NDArray[np.timedelta64])  # type: ignore[assert-type]
assert_type(np.trapezoid(AR_O), npt.NDArray[np.object_] | Any)
assert_type(np.trapezoid(AR_O, AR_LIKE_f8), npt.NDArray[np.object_] | Any)

# meshgrid
assert_type(np.meshgrid(), tuple[()])
assert_type(
    np.meshgrid(AR_f8),
    tuple[
        np.ndarray[tuple[int], np.dtype[np.float64]],
    ],
)
assert_type(
    np.meshgrid(AR_c16, indexing="ij"),
    tuple[
        np.ndarray[tuple[int], np.dtype[np.complex128]],
    ],
)
assert_type(
    np.meshgrid(AR_i8, AR_f8, copy=False),
    tuple[
        np.ndarray[tuple[int, int], np.dtype[np.int64]],
        np.ndarray[tuple[int, int], np.dtype[np.float64]],
    ],
)
assert_type(
    np.meshgrid(AR_LIKE_f8, AR_f8),
    tuple[
        np.ndarray[tuple[int, int]],
        np.ndarray[tuple[int, int], np.dtype[np.float64]],
    ],
)
assert_type(
    np.meshgrid(AR_f8, AR_LIKE_f8),
    tuple[
        np.ndarray[tuple[int, int], np.dtype[np.float64]],
        np.ndarray[tuple[int, int]],
    ],
)
assert_type(
    np.meshgrid(AR_LIKE_f8, AR_LIKE_f8),
    tuple[
        np.ndarray[tuple[int, int]],
        np.ndarray[tuple[int, int]],
    ],
)
assert_type(
    np.meshgrid(AR_f8, AR_i8, AR_c16),
    tuple[
        np.ndarray[tuple[int, int, int], np.dtype[np.float64]],
        np.ndarray[tuple[int, int, int], np.dtype[np.int64]],
        np.ndarray[tuple[int, int, int], np.dtype[np.complex128]],
    ],
)
assert_type(np.meshgrid(AR_f8, AR_f8, AR_f8, AR_f8), tuple[npt.NDArray[np.float64], ...])
assert_type(np.meshgrid(AR_f8, AR_f8, AR_f8, AR_LIKE_f8), tuple[np.ndarray, ...])
assert_type(np.meshgrid(*AR_LIKE_f8), tuple[np.ndarray, ...])

# delete
assert_type(np.delete(AR_f8, np.s_[:5]), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.delete(AR_LIKE_f8, [0, 4, 9], axis=0), np.ndarray)

# insert
assert_type(np.insert(AR_f8, np.s_[:5], 5), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.insert(AR_LIKE_f8, [0, 4, 9], [0.5, 9.2, 7], axis=0), np.ndarray)

# append
assert_type(np.append(f8, f8), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.append(AR_f8, AR_f8), np.ndarray[tuple[int], np.dtype[np.float64]])
assert_type(np.append(AR_LIKE_f8, AR_LIKE_c16, axis=0), np.ndarray)
assert_type(np.append(AR_f8, AR_LIKE_f8, axis=0), np.ndarray)

# digitize
assert_type(np.digitize(4.5, [1]), np.intp)
assert_type(np.digitize(AR_f8, [1, 2, 3]), npt.NDArray[np.intp])



"""Test that all ufuncs have float32-preserving signatures.

This was once guaranteed through the code generation script for
generating ufuncs, `scipy/special/_generate_pyx.py`. Starting with
gh-20260, SciPy developers have begun moving to generate ufuncs
through direct use of the NumPy C API (through C++). Existence of
float32 preserving signatures must now be tested since it is no
longer guaranteed.
"""

import numpy as np
import pytest
import scipy.special._ufuncs
import scipy.special._gufuncs
import scipy.special as special

from numpy.testing import assert_equal


# Single precision is not implemented for these ufuncs;
# floating point inputs must be float64.
exceptions = ['_gen_harmonic', '_normalized_gen_harmonic']

_ufuncs = []
for funcname in dir(scipy.special._ufuncs):
    if funcname not in exceptions:
        _ufuncs.append(getattr(scipy.special._ufuncs, funcname))
for funcname in dir(scipy.special._gufuncs):
    _ufuncs.append(getattr(scipy.special._gufuncs, funcname))

# Not all module members are actually ufuncs
_ufuncs = [func for func in _ufuncs if isinstance(func, np.ufunc)]

@pytest.mark.parametrize("ufunc", _ufuncs)
def test_ufunc_signatures(ufunc):

    # From _generate_pyx.py
    # "Don't add float32 versions of ufuncs with integer arguments, as this
    # can lead to incorrect dtype selection if the integer arguments are
    # arrays, but float arguments are scalars.
    # This may be a NumPy bug, but we need to work around it.
    # cf. gh-4895, https://github.com/numpy/numpy/issues/5895"
    types = set(sig for sig in ufunc.types
                if not ("l" in sig or "i" in sig or "q" in sig or "p" in sig))

    # Generate the full expanded set of signatures which should exist. There
    # should be matching float and double versions of any existing signature.
    expanded_types = set()
    for sig in types:
        expanded_types.update(
            [sig.replace("d", "f").replace("D", "F"),
             sig.replace("f", "d").replace("F", "D")]
        )
    assert types == expanded_types


def _get_nan_val(typecode):
    return {
        "f": np.asarray(np.nan, dtype=np.float32),
        "d": np.asarray(np.nan, dtype=np.float64),
        "g": np.asarray(np.nan, dtype=np.longdouble),
        "F": np.asarray(np.nan + np.nan*1j, dtype=np.complex64),
        "D": np.asarray(np.nan + np.nan*1j, dtype=np.complex128),
        "G": np.asarray(np.nan + np.nan*1j, dtype=np.clongdouble),
    }[typecode]

skips = {
    # These hit https://github.com/scipy/xsf/issues/94
    "eval_chebyc",
    "eval_chebys",
    "eval_chebyt",
    "eval_chebyu",
    "eval_gegenbauer",
    "eval_genlaguerre",
    "eval_jacobi",
    "eval_laguerre",
    "eval_legendre",
    "eval_sh_chebyt",
    "eval_sh_chebyu",
    "eval_sh_jacobi",
    "eval_sh_legendre",
    "hyp1f1",
    "hyp2f1",
    # warns if called with noninteger values of some args
    "bdtr",
    "bdtrc",
    "bdtri",
}
_multi_arg_public_ufuncs = [
    ufunc for ufunc in _ufuncs if ufunc.nargs - ufunc.nout > 1
    and ufunc.__name__ not in skips and ufunc.__name__ in special.__all__
]


@pytest.mark.parametrize("ufunc", _multi_arg_public_ufuncs)
def test_nep50(ufunc):
    # Test that functions with multiple arguments respect nep50 promotion rules.
    rng = np.random.default_rng(1234)
    # As in test_ufunc_signatures, filter out signatures involving integers.
    types = set(sig for sig in ufunc.types
                if not ("l" in sig or "i" in sig or "q" in sig or "p" in sig))
    for sig in types:
        input_types, output_types = sig.split("->")
        # since we only care about dtypes and not values here, just use an appropriately
        # typed nan for each argument.
        args = [_get_nan_val(typecode) for typecode in input_types]
        # swap out a random one of the nans with an appropriately typed numpy scalar.
        idx = rng.choice(len(args))
        args[idx] = float("nan") if np.isrealobj(args[idx]) else complex("nan")
        result = ufunc(*args)
        result = [result] if len(output_types) == 1 else result
        result = np.asarray(result)
        
        # Test that the output is an appropriately typed nan. This also implicitly
        # tests that ufuncs propagate nans correctly.
        assert_equal(
            result, np.asarray([_get_nan_val(typecode) for typecode in output_types]),
            strict=True
        )



import sys
from fontTools.ttLib import OPTIMIZE_FONT_SPEED, TTLibError, TTLibFileIsCollectionError
from fontTools.ttLib.ttFont import *
from fontTools.ttLib.ttCollection import TTCollection


def main(args=None):
    """Open/save fonts with TTFont() or TTCollection()

      ./fonttools ttLib [-oFILE] [-yNUMBER] files...

    If multiple files are given on the command-line,
    they are each opened (as a font or collection),
    and added to the font list.

    If -o (output-file) argument is given, the font
    list is then saved to the output file, either as
    a single font, if there is only one font, or as
    a collection otherwise.

    If -y (font-number) argument is given, only the
    specified font from collections is opened.

    The above allow extracting a single font from a
    collection, or combining multiple fonts into a
    collection.

    If --lazy or --no-lazy are give, those are passed
    to the TTFont() or TTCollection() constructors.
    """
    from fontTools import configLogger

    if args is None:
        args = sys.argv[1:]

    import argparse

    parser = argparse.ArgumentParser(
        "fonttools ttLib",
        description="Open/save fonts with TTFont() or TTCollection()",
        epilog="""
		If multiple files are given on the command-line,
		they are each opened (as a font or collection),
		and added to the font list.

		The above, when combined with -o / --output,
		allows for extracting a single font from a
		collection, or combining multiple fonts into a
		collection.
		""",
    )
    parser.add_argument("font", metavar="font", nargs="*", help="Font file.")
    parser.add_argument(
        "-t", "--table", metavar="table", action="append", help="Tables to decompile."
    )
    parser.add_argument(
        "-o", "--output", metavar="FILE", default=None, help="Output file."
    )
    parser.add_argument(
        "-y", metavar="NUMBER", default=-1, help="Font number to load from collections."
    )
    parser.add_argument(
        "--lazy", action="store_true", default=None, help="Load fonts lazily."
    )
    parser.add_argument(
        "--no-lazy", dest="lazy", action="store_false", help="Load fonts immediately."
    )
    parser.add_argument(
        "--flavor",
        dest="flavor",
        default=None,
        help="Flavor of output font. 'woff' or 'woff2'.",
    )
    parser.add_argument(
        "--no-recalc-timestamp",
        dest="recalcTimestamp",
        action="store_false",
        help="Keep the original font 'modified' timestamp.",
    )
    parser.add_argument(
        "-b",
        dest="recalcBBoxes",
        action="store_false",
        help="Don't recalc glyph bounding boxes: use the values in the original font.",
    )
    parser.add_argument(
        "--optimize-font-speed",
        action="store_true",
        help=(
            "Enable optimizations that prioritize speed over file size. This "
            "mainly affects how glyf table and gvar / VARC tables are compiled."
        ),
    )
    options = parser.parse_args(args)

    fontNumber = int(options.y) if options.y is not None else None
    outFile = options.output
    lazy = options.lazy
    flavor = options.flavor
    tables = options.table
    recalcBBoxes = options.recalcBBoxes
    recalcTimestamp = options.recalcTimestamp
    optimizeFontSpeed = options.optimize_font_speed

    fonts = []
    for f in options.font:
        try:
            font = TTFont(
                f,
                recalcBBoxes=recalcBBoxes,
                recalcTimestamp=recalcTimestamp,
                fontNumber=fontNumber,
                lazy=lazy,
            )
            if optimizeFontSpeed:
                font.cfg[OPTIMIZE_FONT_SPEED] = optimizeFontSpeed
            fonts.append(font)
        except TTLibFileIsCollectionError:
            collection = TTCollection(f, lazy=lazy)
            fonts.extend(collection.fonts)

    if tables is None:
        if lazy is False:
            tables = ["*"]
        elif optimizeFontSpeed:
            tables = {"glyf", "gvar", "VARC"}.intersection(font.keys())
        else:
            tables = []
    for font in fonts:
        if "GlyphOrder" in tables:
            font.getGlyphOrder()
        for table in tables if "*" not in tables else font.keys():
            font[table]  # Decompiles

    if outFile is not None:
        if len(fonts) == 1:
            fonts[0].flavor = flavor
            fonts[0].save(outFile)
        else:
            if flavor is not None:
                raise TTLibError("Cannot set flavor for collections.")
            collection = TTCollection()
            collection.fonts = fonts
            collection.save(outFile)


if __name__ == "__main__":
    sys.exit(main())



import os
import tempfile

import pytest

from . import util


# Note: do not mark as slow! This is one of a small set of f2py compile tests retained
# in the default suite invocation for compile-path coverage.
class TestAssumedShapeSumExample(util.F2PyTest):
    sources = [
        util.getpath("tests", "src", "assumed_shape", "foo_free.f90"),
        util.getpath("tests", "src", "assumed_shape", "foo_use.f90"),
        util.getpath("tests", "src", "assumed_shape", "precision.f90"),
        util.getpath("tests", "src", "assumed_shape", "foo_mod.f90"),
        util.getpath("tests", "src", "assumed_shape", ".f2py_f2cmap"),
    ]

    def test_all(self):
        r = self.module.fsum([1, 2])
        assert r == 3
        r = self.module.sum([1, 2])
        assert r == 3
        r = self.module.sum_with_use([1, 2])
        assert r == 3

        r = self.module.mod.sum([1, 2])
        assert r == 3
        r = self.module.mod.fsum([1, 2])
        assert r == 3


@pytest.mark.slow
class TestF2cmapOption(TestAssumedShapeSumExample):
    def setup_method(self):
        # Use a custom file name for .f2py_f2cmap
        self.sources = list(self.sources)
        f2cmap_src = self.sources.pop(-1)

        self.f2cmap_file = tempfile.NamedTemporaryFile(delete=False)
        with open(f2cmap_src, "rb") as f:
            self.f2cmap_file.write(f.read())
        self.f2cmap_file.close()

        self.sources.append(self.f2cmap_file.name)
        self.options = ["--f2cmap", self.f2cmap_file.name]

        super().setup_method()

    def teardown_method(self):
        os.unlink(self.f2cmap_file.name)



from collections.abc import Generator
import contextlib

from matplotlib import RcParams
from matplotlib.typing import RcStyleType

USER_LIBRARY_PATHS: list[str] = ...

def use(style: RcStyleType) -> None: ...
@contextlib.contextmanager
def context(
    style: RcStyleType, after_reset: bool = ...
) -> Generator[None, None, None]: ...

library: dict[str, RcParams]
available: list[str]

def reload_library() -> None: ...

__all__ = ['use', 'context', 'available', 'library', 'reload_library']



from typing import Any

import numpy as np

# NOTE: `_StrLike_co` and `_BytesLike_co` are pointless, as `np.str_` and
# `np.bytes_` are already subclasses of their builtin counterpart
type _CharLike_co = str | bytes

# The `<X>Like_co` type-aliases below represent all scalars that can be
# coerced into `<X>` (with the casting rule `same_kind`)
type _BoolLike_co = bool | np.bool
type _UIntLike_co = bool | np.unsignedinteger | np.bool
type _IntLike_co = int | np.integer | np.bool
type _FloatLike_co = float | np.floating | np.integer | np.bool
type _ComplexLike_co = complex | np.number | np.bool
type _NumberLike_co = _ComplexLike_co
type _TD64Like_co = int | np.timedelta64 | np.integer | np.bool
# `_VoidLike_co` is technically not a scalar, but it's close enough
type _VoidLike_co = tuple[Any, ...] | np.void
type _ScalarLike_co = complex | str | bytes | np.generic



import numpy as np
import pytest
from scipy import stats

from scipy._lib._array_api import xp_assert_close, xp_assert_equal, _count_nonmasked
from scipy._lib._array_api import make_xp_pytest_param, make_xp_test_case
from scipy._lib._array_api import SCIPY_ARRAY_API, is_torch
from scipy.stats._stats_py import _xp_mean, _xp_var


marray = pytest.importorskip('marray')
pytestmark = [
    pytest.mark.skipif(
        not SCIPY_ARRAY_API,
        reason=(
            "special function dispatch to marray required for these tests"
            " is hidden behind SCIPY_ARRAY_API flag."
        ),
    ),
]

skip_backend = pytest.mark.skip_xp_backends

def get_arrays(n_arrays, *, dtype='float64', xp=np, shape=(7, 8), all_unique=True,
               seed=84912165484321):
    mxp = marray._get_namespace(xp)
    rng = np.random.default_rng(seed)

    datas, masks = [], []
    for i in range(n_arrays):
        data = (rng.random(size=shape) if all_unique
                else rng.integers(np.min(shape) // 2, size=shape))
        if dtype.startswith('complex'):
            data = 10*data * 10j*rng.standard_normal(size=shape)
        data = data.astype(dtype)
        datas.append(data)
        mask = rng.random(size=shape) > 0.75
        masks.append(mask)

    marrays = []
    nan_arrays = []
    for array, mask in zip(datas, masks):
        marrays.append(mxp.asarray(array, mask=mask))
        nan_array = array.copy()
        nan_array[mask] = xp.nan
        nan_arrays.append(nan_array)

    return mxp, marrays, nan_arrays


@skip_backend('jax.numpy', reason="JAX doesn't allow item assignment.")
@pytest.mark.parametrize('fun, kwargs', [
    make_xp_pytest_param(stats.gmean, {}),
    make_xp_pytest_param(stats.hmean, {}),
    make_xp_pytest_param(stats.pmean, {'p': 2}),
    make_xp_pytest_param(stats.expectile, {'alpha': 0.4}),
    make_xp_pytest_param(_xp_mean, {}),
])
@pytest.mark.parametrize('weighted', [False, True])
@pytest.mark.parametrize('axis', [0, 1, None])
def test_one_sample_weighted_reducing(fun, kwargs, weighted, axis, xp):
    mxp, marrays, narrays = get_arrays(2, xp=xp)
    mweights = marrays[1] if weighted else None
    nweights = narrays[1] if weighted else None
    res = fun(marrays[0], weights=mweights, axis=axis, **kwargs)
    ref = fun(narrays[0], weights=nweights, nan_policy='omit', axis=axis, **kwargs)
    xp_assert_close(res.data, xp.asarray(ref))


@skip_backend('dask.array', reason='Arrays need `device` attribute: dask/dask#11711')
@skip_backend('jax.numpy', reason="JAX doesn't allow item assignment.")
@pytest.mark.parametrize('fun, kwargs',
    [make_xp_pytest_param(stats.moment, {'order': 2}),
     make_xp_pytest_param(stats.moment, {'center': 0.5}),
     make_xp_pytest_param(stats.skew, {}),
     make_xp_pytest_param(stats.skew, {'bias': False}),
     make_xp_pytest_param(stats.kurtosis, {}),
     make_xp_pytest_param(stats.kurtosis, {'fisher': False, 'bias': False}),
     make_xp_pytest_param(stats.sem, {}),
     make_xp_pytest_param(stats.sem, {'ddof': 2}),
     make_xp_pytest_param(stats.kstat, {'n': 1}),
     make_xp_pytest_param(stats.kstat, {'n': 2}),
     make_xp_pytest_param(stats.kstat, {'n': 3}),
     make_xp_pytest_param(stats.kstat, {'n': 4}),
     make_xp_pytest_param(stats.kstatvar, {'n': 1}),
     make_xp_pytest_param(stats.kstatvar, {'n': 2}),
     make_xp_pytest_param(stats.circmean, {}),
     make_xp_pytest_param(stats.circmean, {'low': 0, 'high': 1}),
     make_xp_pytest_param(stats.circvar, {}),
     make_xp_pytest_param(stats.circvar, {'low': 0, 'high': 1}),
     make_xp_pytest_param(stats.circstd, {}),
     make_xp_pytest_param(stats.circstd, {'low': 0, 'high': 1, 'normalize':True}),
     make_xp_pytest_param(stats.gstd, {}),
     make_xp_pytest_param(stats.gstd, {'ddof': 2}),
     make_xp_pytest_param(_xp_var, {}),
     make_xp_pytest_param(_xp_var, {'correction': 1, 'keepdims': True}),
     make_xp_pytest_param(stats.variation, {}),
     make_xp_pytest_param(stats.variation, {'ddof': 1}),
     make_xp_pytest_param(stats.tmean, dict(limits=(0.1, 0.9))),
     make_xp_pytest_param(stats.tvar, dict(limits=(0.1, 0.9), inclusive=(False, True))),
     make_xp_pytest_param(stats.tstd, dict(limits=(0.1, 0.9), inclusive=(True, False))),
     make_xp_pytest_param(stats.tsem, dict(limits=(0.1, 0.9), inclusive=(False,)*2)),
     make_xp_pytest_param(stats.tmin, dict(lowerlimit=0.5, inclusive=True)),
     make_xp_pytest_param(stats.tmax, dict(upperlimit=0.5, inclusive=False)),
     make_xp_pytest_param(stats.iqr, {'interpolation': 'nearest'}),
     make_xp_pytest_param(stats.iqr, {'rng': (10, 90), 'scale': 'normal'}),
     make_xp_pytest_param(stats.median_abs_deviation, {}),
     make_xp_pytest_param(stats.median_abs_deviation, {'scale': 'normal'}),
     make_xp_pytest_param(stats.trim_mean, {'proportiontocut': 0.1}),
     ])
@pytest.mark.parametrize('axis', [0, 1, None])
def test_one_sample_reducing(fun, kwargs, axis, xp):
    mxp, marrays, narrays = get_arrays(1, xp=xp)
    kwargs = dict(axis=axis) | kwargs
    res = fun(marrays[0], **kwargs)
    ref = fun(narrays[0], nan_policy='omit', **kwargs)
    xp_assert_close(res.data, xp.asarray(ref))


@make_xp_test_case(stats.mode)
@pytest.mark.parametrize('dtype', ['int32', 'float64'])
@pytest.mark.parametrize('shape', [10, (7, 8)])
@pytest.mark.parametrize('axis', [0, 1, None])
def test_mode(dtype, shape, axis, xp):
    mxp, marrays, narrays = get_arrays(1, shape=shape, all_unique=False, xp=xp)
    res = stats.mode(mxp.astype(marrays[0], getattr(mxp, dtype)))
    ref = stats.mode(*narrays, nan_policy='omit')
    xp_assert_close(res.mode.data, xp.asarray(ref.mode.astype(dtype)))
    xp_assert_close(res.count.data, xp.asarray(ref.count))


@make_xp_test_case(stats.describe)
@skip_backend('dask.array', reason='Arrays need `device` attribute: dask/dask#11711')
@skip_backend('jax.numpy', reason="JAX doesn't allow item assignment.")
@pytest.mark.parametrize('axis', [0, 1, None])
@pytest.mark.parametrize('kwargs', [{}, {'bias': False}])
def test_describe(axis, kwargs, xp):
    mxp, marrays, narrays = get_arrays(1, xp=xp)
    kwargs = dict(axis=axis) | kwargs
    res = stats.describe(marrays[0], **kwargs)
    ref = stats.describe(narrays[0], nan_policy='omit', **kwargs)
    xp_assert_close(res.nobs.data, xp.asarray(ref.nobs))
    xp_assert_close(res.minmax[0].data, xp.asarray(ref.minmax[0].data))
    xp_assert_close(res.minmax[1].data, xp.asarray(ref.minmax[1].data))
    # copy reference arrays due to torch complaint about non-writeable buffer
    xp_assert_close(res.variance.data,
                    xp.asarray(np.asarray(ref.variance.data, copy=True)))
    xp_assert_close(res.skewness.data,
                    xp.asarray(np.asarray(ref.skewness.data, copy=True)))
    xp_assert_close(res.kurtosis.data,
                    xp.asarray(np.asarray(ref.kurtosis.data, copy=True)))


@skip_backend('dask.array', reason='shape (nan,) in _xp_var when ddof=1')
@skip_backend('jax.numpy', reason="JAX doesn't allow item assignment.")
@pytest.mark.parametrize('fun', [make_xp_pytest_param(stats.zscore),
                                 make_xp_pytest_param(stats.gzscore),
                                 make_xp_pytest_param(stats.zmap)])
@pytest.mark.parametrize('ddof', [0, 1])
@pytest.mark.parametrize('axis', [0, 1, None])
def test_one_sample_non_reducing(fun, ddof, axis, xp):
    mxp, marrays, narrays = (get_arrays(2, xp=xp) if fun == stats.zmap
                             else get_arrays(1, xp=xp))
    res = fun(*marrays, ddof=ddof, axis=axis)
    ref = xp.asarray(fun(*narrays, ddof=ddof, nan_policy='omit', axis=axis))
    xp_assert_close(res.data[~res.mask], ref[~xp.isnan(ref)])
    xp_assert_equal(res.mask, marrays[0].mask)


@skip_backend('dask.array', reason='Arrays need `device` attribute: dask/dask#11711')
@skip_backend('jax.numpy', reason="JAX doesn't allow item assignment.")
@pytest.mark.parametrize('f', [make_xp_pytest_param(stats.ttest_1samp),
                               make_xp_pytest_param(stats.ttest_rel),
                               make_xp_pytest_param(stats.ttest_ind)])
@pytest.mark.parametrize('alternative', ['less', 'greater', 'two-sided'])
@pytest.mark.parametrize('axis', [0, 1, None])
def test_ttests(f, alternative, axis, xp):
    f_name = f.__name__
    mxp, marrays, narrays = get_arrays(2, xp=xp)
    if f_name == 'ttest_1samp':
        marrays[1] = mxp.mean(marrays[1], axis=axis, keepdims=axis is not None)
        narrays[1] = np.nanmean(narrays[1], axis=axis, keepdims=axis is not None)
    res = f(*marrays, alternative=alternative, axis=axis)
    ref = f(*narrays, alternative=alternative, nan_policy='omit', axis=axis)
    xp_assert_close(res.statistic.data, xp.asarray(ref.statistic))
    xp_assert_close(res.pvalue.data, xp.asarray(ref.pvalue))
    res_ci = res.confidence_interval()
    ref_ci = ref.confidence_interval()
    xp_assert_close(res_ci.low.data, xp.asarray(ref_ci.low))
    xp_assert_close(res_ci.high.data, xp.asarray(ref_ci.high))


@skip_backend('dask.array', reason='Arrays need `device` attribute: dask/dask#11711')
@skip_backend('jax.numpy', reason="JAX doesn't allow item assignment.")
@pytest.mark.filterwarnings("ignore::scipy.stats._axis_nan_policy.SmallSampleWarning")
@pytest.mark.parametrize('f, args', [
     make_xp_pytest_param(stats.skewtest, tuple()),
     make_xp_pytest_param(stats.kurtosistest, tuple()),
     make_xp_pytest_param(stats.normaltest, tuple()),
     make_xp_pytest_param(stats.jarque_bera, tuple()),
     make_xp_pytest_param(stats.cramervonmises, (stats.norm.cdf,)),  # type:ignore[attr-defined]
])
@pytest.mark.parametrize('alternative', ['less', 'greater', 'two-sided'])
@pytest.mark.parametrize('axis', [0, 1, None])
def test_goodness_of_fit(f, args, alternative, axis, xp):
    mxp, marrays, narrays = get_arrays(1, xp=xp, shape=(10, 11))

    if f in {stats.skewtest, stats.kurtosistest}:
        kwds = {'alternative': alternative}
    else:
        if alternative != 'greater':
            pytest.skip(f'str({f.__name__} does not support multiple alternatives.')
        kwds = {}

    res = f(*marrays, *args, **kwds, axis=axis)
    ref = f(*narrays, *args, **kwds, nan_policy='omit', axis=axis)

    xp_assert_close(res.statistic.data, xp.asarray(ref.statistic))
    xp_assert_close(res.pvalue.data, xp.asarray(ref.pvalue), atol=1e-15)


@skip_backend('dask.array', reason='Arrays need `device` attribute: dask/dask#11711')
@skip_backend('jax.numpy', reason="JAX doesn't allow item assignment.")
@pytest.mark.parametrize('f, lambda_', [
    make_xp_pytest_param(stats.chisquare, {}),
    make_xp_pytest_param(stats.power_divergence, {'lambda_': 'pearson'}),
    make_xp_pytest_param(stats.power_divergence, {'lambda_': 'log-likelihood'}),
    make_xp_pytest_param(stats.power_divergence, {'lambda_': 'freeman-tukey'}),
    make_xp_pytest_param(stats.power_divergence,
                         {'lambda_': 'mod-log-likelihood'}),
    make_xp_pytest_param(stats.power_divergence, {'lambda_': 'neyman'}),
    make_xp_pytest_param(stats.power_divergence, {'lambda_': 'cressie-read'}),
])
@pytest.mark.parametrize('ddof', [0, 1])
@pytest.mark.parametrize('axis', [0, 1, None])
def test_power_divergence_chisquare(f, lambda_, ddof, axis, xp):
    mxp, marrays, narrays = get_arrays(2, xp=xp, shape=(5, 6))

    kwargs = dict(axis=axis, ddof=ddof)

    # test 1-arg
    res = f(marrays[0], **lambda_, **kwargs)
    ref = stats.power_divergence(narrays[0], nan_policy='omit', **lambda_, **kwargs)

    xp_assert_close(res.statistic.data, xp.asarray(ref[0]))
    xp_assert_close(res.pvalue.data, xp.asarray(ref[1]))

    # test 2-arg
    common_mask = np.isnan(narrays[0]) | np.isnan(narrays[1])
    normalize = (np.nansum(narrays[1] * ~common_mask, axis=axis, keepdims=True)
                 / np.nansum(narrays[0] * ~common_mask, axis=axis, keepdims=True))
    marrays[0] *= xp.asarray(normalize)
    narrays[0] *= normalize

    res = f(*marrays, **lambda_, **kwargs)
    ref = stats.power_divergence(*narrays, nan_policy='omit', **lambda_, **kwargs)

    xp_assert_close(res.statistic.data, xp.asarray(ref[0]))
    xp_assert_close(res.pvalue.data, xp.asarray(ref[1]))


@make_xp_test_case(stats.combine_pvalues)
@skip_backend('jax.numpy', reason="JAX doesn't allow item assignment.")
@pytest.mark.parametrize('method', ['fisher', 'pearson', 'mudholkar_george',
                                    'tippett', 'stouffer'])
@pytest.mark.parametrize('axis', [0, 1, None])
def test_combine_pvalues(method, axis, xp):
    mxp, marrays, narrays = get_arrays(2, xp=xp, shape=(10, 11))

    kwargs = dict(method=method, axis=axis)
    res = stats.combine_pvalues(marrays[0], **kwargs)
    ref = stats.combine_pvalues(narrays[0], nan_policy='omit', **kwargs)

    xp_assert_close(res.statistic.data, xp.asarray(ref.statistic))
    xp_assert_close(res.pvalue.data, xp.asarray(ref.pvalue))

    if method != 'stouffer':
        return

    # test method='stouffer' with weights
    res = stats.combine_pvalues(marrays[0], weights=marrays[1], **kwargs)
    ref = stats.combine_pvalues(narrays[0], weights=narrays[1],
                                nan_policy='omit', **kwargs)

    xp_assert_close(res.statistic.data, xp.asarray(ref.statistic))
    xp_assert_close(res.pvalue.data, xp.asarray(ref.pvalue))


@make_xp_test_case(stats.ttest_ind_from_stats)
@skip_backend('jax.numpy', reason="JAX doesn't allow item assignment.")
@pytest.mark.parametrize("alternative", ['less', 'greater', 'two-sided'])
@pytest.mark.parametrize("equal_var", [False, True])
def test_ttest_ind_from_stats(alternative, equal_var, xp):
    shape = (10, 11)
    mxp, marrays, narrays = get_arrays(6, xp=xp, shape=shape)
    mask = np.sum(np.stack([np.isnan(arg) for arg in narrays]), axis=0).astype(bool)
    narrays = [arg[~mask] for arg in narrays]
    marrays[2], marrays[5] = marrays[2] * 100, marrays[5] * 100
    narrays[2], narrays[5] = narrays[2] * 100, narrays[5] * 100

    kwargs = dict(alternative=alternative, equal_var=equal_var)
    res = stats.ttest_ind_from_stats(*marrays, **kwargs)
    ref = stats.ttest_ind_from_stats(*narrays, **kwargs)

    mask = xp.asarray(mask)
    assert xp.any(mask) and xp.any(~mask)
    xp_assert_close(res.statistic.data[~mask], xp.asarray(ref.statistic))
    xp_assert_close(res.pvalue.data[~mask], xp.asarray(ref.pvalue))
    xp_assert_close(res.statistic.mask, mask)
    xp_assert_close(res.pvalue.mask, mask)
    assert res.statistic.shape == shape
    assert res.pvalue.shape == shape


def test_length_nonmasked_marray_iterable_axis_raises():
    xp = marray._get_namespace(np)

    data = [[1.0, 2.0], [3.0, 4.0]]
    mask = [[False, False], [True, False]]
    marr = xp.asarray(data, mask=mask)

    # Axis tuples are not currently supported for MArray input.
    # This test can be removed after support is added.
    with pytest.raises(NotImplementedError,
        match="`axis` must be an integer or None for use with `MArray`"):
        _count_nonmasked(marr, axis=(0, 1), xp=xp)


@make_xp_test_case(stats.directional_stats)
@skip_backend('jax.numpy', reason="JAX doesn't allow item assignment.")
@pytest.mark.filterwarnings("ignore::RuntimeWarning")  # mdhaber/marray#120
@pytest.mark.parametrize('normalize', [False, True])
@pytest.mark.parametrize('axis', [0, 1])
def test_directional_stats(normalize, axis, xp):
    mxp, marrays, narrays = get_arrays(1, shape=(19, 20, 3), xp=xp)
    res = stats.directional_stats(*marrays, normalize=normalize, axis=axis)

    x, = narrays
    if axis == 0:
        x = np.swapaxes(x, 0, 1)

    for i in range(x.shape[0]):
        xi = x[i]
        xi = xi[~np.any(np.isnan(xi), axis=-1), :]
        ref = stats.directional_stats(xi, normalize=normalize)
        xp_assert_close(res.mean_direction.data[i, ...],
                        xp.asarray(ref.mean_direction))
        xp_assert_close(res.mean_resultant_length.data[i],
                        xp.asarray(ref.mean_resultant_length)[()])
    assert not xp.any(res.mean_direction.mask)
    assert not xp.any(res.mean_resultant_length.mask)


@make_xp_test_case(stats.wilcoxon)
@skip_backend('jax.numpy', reason="JAX doesn't allow item assignment.")
@pytest.mark.parametrize('n_samples', [1, 2])
@pytest.mark.parametrize('zero_method', ['zsplit'])  # TODO: add 'wilcox', 'pratt'
@pytest.mark.parametrize('correction', [False, True])
@pytest.mark.parametrize('alternative', ['less', 'greater', 'two-sided'])
@pytest.mark.parametrize('method', ['asymptotic'])  # TODO: add 'exact', 'auto'
@pytest.mark.parametrize('axis', [0, 1, None])
def test_wilcoxon(n_samples, zero_method, correction, alternative, method, axis, xp):
    mxp, marrays, narrays = get_arrays(n_samples, xp=xp, seed=84912165484322)
    kwargs = dict(zero_method=zero_method, correction=correction,
                  alternative=alternative, method=method)
    res = stats.wilcoxon(*marrays, axis=axis, **kwargs)
    ref = stats.wilcoxon(*narrays, nan_policy='omit', axis=axis, **kwargs)
    xp_assert_close(res.statistic.data, xp.asarray(ref.statistic))
    xp_assert_close(res.pvalue.data, xp.asarray(ref.pvalue))


@make_xp_test_case(stats.mannwhitneyu)
@skip_backend('jax.numpy', reason="JAX doesn't allow item assignment.")
@pytest.mark.parametrize('use_continuity', [False, True])
@pytest.mark.parametrize('alternative', ['less', 'greater', 'two-sided'])
@pytest.mark.parametrize('method', ['asymptotic'])  # TODO: add 'exact', 'auto'
@pytest.mark.parametrize('axis', [0, 1, None])
def test_mannwhitneyu(use_continuity, alternative, method, axis, xp):
    mxp, marrays, narrays = get_arrays(2, xp=xp, seed=84912165484322)
    kwargs = dict(use_continuity=use_continuity, alternative=alternative, method=method)
    res = stats.mannwhitneyu(*marrays, axis=axis, **kwargs)
    ref = stats.mannwhitneyu(*narrays, nan_policy='omit', axis=axis, **kwargs)
    xp_assert_close(res.statistic.data, xp.asarray(ref.statistic))
    xp_assert_close(res.pvalue.data, xp.asarray(ref.pvalue))


@skip_backend('jax.numpy', reason="JAX doesn't allow item assignment.")
@pytest.mark.parametrize('fun', [make_xp_pytest_param(stats.ks_1samp),
                                 make_xp_pytest_param(stats.kstest)])
@pytest.mark.parametrize('method', ['exact', 'approx', 'asymp'])  # auto == exact
@pytest.mark.parametrize('alternative', ['less', 'greater', 'two-sided'])
@pytest.mark.parametrize('axis', [0, 1, None])
def test_ks_1samp(fun, method, alternative, axis, xp):
    mxp, marrays, narrays = get_arrays(1, xp=xp, seed=84912165484322)
    kwargs = dict(method=method, alternative=alternative, axis=axis)
    res = fun(*marrays, stats.norm.cdf, **kwargs)
    ref = stats.ks_1samp(*narrays, stats.norm.cdf, nan_policy='omit', **kwargs)
    xp_assert_close(res.statistic.data, xp.asarray(ref.statistic))
    xp_assert_close(res.pvalue.data, xp.asarray(ref.pvalue))
    xp_assert_equal(res.statistic_location.data, xp.asarray(ref.statistic_location))
    xp_assert_equal(res.statistic_sign.data,
                    xp.asarray(ref.statistic_sign, dtype=xp.int8))


@skip_backend('jax.numpy', reason="JAX doesn't allow item assignment.")
@pytest.mark.parametrize('fun, kwargs', [
    make_xp_pytest_param(stats.brunnermunzel, {}),
    make_xp_pytest_param(stats.brunnermunzel, {'distribution': 'normal'}),
    make_xp_pytest_param(stats.cramervonmises_2samp, {'method': 'exact'}),
    # make_xp_pytest_param(stats.cramervonmises_2samp, {}),  # TODO: add methods
])
@pytest.mark.parametrize('axis', [0, 1, None])
def test_two_sample_tests(fun, kwargs, axis, xp):
    if fun == stats.cramervonmises_2samp and axis is None:
        pytest.skip("Sample too large for exact method.")
    mxp, marrays, narrays = get_arrays(2, xp=xp, seed=84912165484322)
    res = fun(*marrays, axis=axis, **kwargs)
    ref = fun(*narrays, nan_policy='omit', axis=axis, **kwargs)
    xp_assert_close(res.statistic.data, xp.asarray(ref.statistic))
    xp_assert_close(res.pvalue.data, xp.asarray(ref.pvalue))


@make_xp_test_case(stats.ks_2samp)
@skip_backend('jax.numpy', reason="JAX doesn't allow item assignment.")
@pytest.mark.parametrize('fun', [make_xp_pytest_param(stats.ks_2samp),
                                 make_xp_pytest_param(stats.kstest)])
@pytest.mark.parametrize('method', ['exact', 'asymp', 'auto'])  # auto == exact
@pytest.mark.parametrize('alternative', ['less', 'greater', 'two-sided'])
@pytest.mark.parametrize('axis', [0, 1, None])
def test_ks_2samp(fun, method, alternative, axis, xp):
    mxp, marrays, narrays = get_arrays(2, xp=xp, seed=84912165484322)
    kwargs = dict(method=method, alternative=alternative, axis=axis)
    res = fun(*marrays, **kwargs)
    ref = stats.ks_2samp(*narrays, nan_policy='omit', **kwargs)
    xp_assert_close(res.statistic.data, xp.asarray(ref.statistic))
    xp_assert_close(res.pvalue.data, xp.asarray(ref.pvalue))
    # with this random data, there often multiple locations where the statistic assumes
    # the most extreme value, so we can't expect these to always match
    # xp_assert_equal(res.statistic_location.data, xp.asarray(ref.statistic_location))
    xp_assert_equal(res.statistic_sign.data,
                    xp.asarray(ref.statistic_sign, dtype=xp.int8))


@make_xp_test_case(stats.kendalltau)
class TestKendallTau:
    @pytest.mark.parametrize('method', ['asymptotic', 'exact'])
    @pytest.mark.parametrize('variant', ['b', 'c'])
    @pytest.mark.parametrize('alternative', ['less', 'greater', 'two-sided'])
    @pytest.mark.parametrize('axis', [0, 1, None])
    def test_omit_masked_elements(self, method, variant, alternative, axis, xp):
        mxp, marrays, narrays = get_arrays(2, shape=(17, 18),
                                           xp=xp, seed=84912165484322)
        kwargs = dict(method=method, variant=variant, alternative=alternative)
        res = stats.kendalltau(*marrays, **kwargs, axis=axis)
        ref = stats.kendalltau(*narrays, **kwargs, nan_policy='omit', axis=axis)
        xp_assert_close(res.statistic.data, xp.asarray(ref.statistic))
        xp_assert_close(res.pvalue.data, xp.asarray(ref.pvalue))

    @pytest.mark.parametrize('nan_policy, message', [
        ('propagate', "`nan_policy='propagate'` is incompatible with MArray input."),
        ('raise', 'The input contains nan values.')
    ])
    def test_input_validation(self, nan_policy, message, xp):
        mxp = marray._get_namespace(xp)
        x = mxp.asarray([1, 2, 3, xp.nan])
        with pytest.raises(ValueError, match=message):
            stats.kendalltau(x, x, nan_policy=nan_policy)


@skip_backend('dask.array', reason='Arrays need `device` attribute: dask/dask#11711')
@skip_backend('jax.numpy', reason="JAX doesn't allow item assignment.")
@pytest.mark.parametrize('fun, kwargs', [
    make_xp_pytest_param(stats.bartlett, {}),
    make_xp_pytest_param(stats.alexandergovern, {}),
    make_xp_pytest_param(stats.levene, {'center': 'median'}),
    make_xp_pytest_param(stats.levene, {'center': 'mean'}),
    make_xp_pytest_param(stats.levene, {'center': 'trimmed'}),
    make_xp_pytest_param(stats.f_oneway, {'equal_var': True}),
    make_xp_pytest_param(stats.f_oneway, {'equal_var': False}),
    make_xp_pytest_param(stats.kruskal, {}),
    make_xp_pytest_param(stats.fligner, {'center': 'median'}),
    make_xp_pytest_param(stats.fligner, {'center': 'mean'}),
    make_xp_pytest_param(stats.fligner, {'center': 'trimmed'}),
])
@pytest.mark.parametrize('axis', [0, 1, None])
def test_k_sample_tests(fun, kwargs, axis, xp):
    mxp, marrays, narrays = get_arrays(3, xp=xp)
    res = fun(*marrays, axis=axis, **kwargs)
    ref = fun(*narrays, nan_policy='omit', axis=axis, **kwargs)
    xp_assert_close(res.statistic.data, xp.asarray(ref.statistic))
    xp_assert_close(res.pvalue.data, xp.asarray(ref.pvalue))


@skip_backend('jax.numpy', reason="JAX currently incompatible with marray")
@pytest.mark.parametrize('fun, kwargs', [
    make_xp_pytest_param(stats.friedmanchisquare, {}),
])
@pytest.mark.parametrize('axis', [0, 1, None])
def test_k_sample_paired_tests(fun, kwargs, axis, xp):
    mxp, marrays, narrays = get_arrays(3, shape=(8, 9), xp=xp)
    res = fun(*marrays, axis=axis, **kwargs)
    ref = fun(*narrays, nan_policy='omit', axis=axis, **kwargs)
    xp_assert_close(res.statistic.data, xp.asarray(ref.statistic))
    xp_assert_close(res.pvalue.data, xp.asarray(ref.pvalue))


@skip_backend('dask.array', reason='Arrays need `device` attribute: dask/dask#11711')
@skip_backend('jax.numpy', reason="JAX doesn't allow item assignment.")
@pytest.mark.parametrize('f', [
    make_xp_pytest_param(stats.pearsonr),
    make_xp_pytest_param(stats.pointbiserialr),
    make_xp_pytest_param(stats.spearmanrho),
])
@pytest.mark.parametrize('alternative', ['less', 'greater', 'two-sided'])
@pytest.mark.parametrize('axis', [0, 1, None])
def test_correlation(f, alternative, axis, xp):
    mxp, marrays, narrays = get_arrays(2, xp=xp)

    kwargs = {} if f == stats.pointbiserialr else {'alternative': alternative}
    res = f(*marrays, **kwargs, axis=axis)

    # `pearsonr` does not have `axis_nan_policy`, so do this manually
    x, y = narrays
    if axis == 0:
        x, y = x.T, y.T
    elif axis is None:
        x, y = x.ravel()[np.newaxis, :], y.ravel()[np.newaxis, :]

    for i in range(x.shape[0]):
        xi, yi = x[i, ...], y[i, ...]
        i = () if axis is None else i

        mask = np.isnan(xi) | np.isnan(yi)
        ref = f(xi[~mask], yi[~mask], **kwargs)

        atol = 1e-7 if (is_torch(xp) and f == stats.spearmanrho) else 0.
        xp_assert_close(res.statistic.data[i], xp.asarray(ref.statistic)[()], atol=atol)
        xp_assert_close(res.pvalue.data[i], xp.asarray(ref.pvalue)[()], atol=atol)

        if f == stats.pearsonr:
            res_ci_low, res_ci_high = res.confidence_interval()
            ref_ci_low, ref_ci_high = ref.confidence_interval()
            xp_assert_close(res_ci_low.data[i], xp.asarray(ref_ci_low)[()])
            xp_assert_close(res_ci_high.data[i], xp.asarray(ref_ci_high)[()])


@skip_backend('jax.numpy', reason="JAX doesn't allow item assignment.")
@skip_backend('array_api_strict', reason="issue with clip.")
@pytest.mark.parametrize('f, method', [
    make_xp_pytest_param(stats.siegelslopes, {'method':'hierarchical'}),
    make_xp_pytest_param(stats.siegelslopes, {'method':'separate'}),
    make_xp_pytest_param(stats.theilslopes, {'method': 'joint'}),
    make_xp_pytest_param(stats.theilslopes, {'method': 'separate'}),
    make_xp_pytest_param(stats.linregress, {'alternative': 'less'}),
    make_xp_pytest_param(stats.linregress, {'alternative': 'greater'}),
    make_xp_pytest_param(stats.linregress, {'alternative': 'two-sided'}),
])
@pytest.mark.parametrize('axis', [0, 1, None])
def test_slopes_intercepts(f, method, axis, xp):
    mxp, marrays, narrays = get_arrays(2, shape=(19, 20), seed=84912165484320, xp=xp)
    res = f(*marrays, axis=axis, **method)
    ref = f(*narrays, nan_policy='omit', axis=axis, **method)

    xp_assert_close(res.slope.data, xp.asarray(ref.slope))
    xp_assert_close(res.intercept.data, xp.asarray(ref.intercept))

    if f == stats.theilslopes:
        xp_assert_close(res.low_slope.data, xp.asarray(ref.low_slope))
        xp_assert_close(res.high_slope.data, xp.asarray(ref.high_slope))
    elif f == stats.linregress:
        xp_assert_close(res.rvalue.data, xp.asarray(ref.rvalue))
        xp_assert_close(res.pvalue.data, xp.asarray(ref.pvalue))
        xp_assert_close(res.stderr.data, xp.asarray(ref.stderr))
        xp_assert_close(res.intercept_stderr.data, xp.asarray(ref.intercept_stderr))


@make_xp_test_case(stats.entropy)
@skip_backend('dask.array', reason="Dask doesn't like the /= when base is not None.")
@skip_backend('jax.numpy', reason="JAX doesn't allow item assignment.")
@pytest.mark.parametrize('qk', [False, True])
@pytest.mark.parametrize('base', [None, 2])
@pytest.mark.parametrize('axis', [0, 1, None])
def test_entropy(qk, base, axis, xp):
    mxp, marrays, narrays = get_arrays(2 if qk else 1, xp=xp)
    res = stats.entropy(*marrays, base=base, axis=axis)
    ref = stats.entropy(*narrays, base=base, nan_policy='omit', axis=axis)
    xp_assert_close(res.data, xp.asarray(ref))


@make_xp_test_case(stats.rankdata)
@skip_backend('jax.numpy', reason="JAX currently incompatible with marray")
@pytest.mark.parametrize('method', ['average', 'min', 'max', 'dense', 'ordinal'])
@pytest.mark.parametrize('axis', [0, 1, None])
def test_rankdata(method, axis, xp):
    mxp, marrays, narrays = get_arrays(1, xp=xp)
    res = stats.rankdata(*marrays, method=method, axis=axis)
    ref = stats.rankdata(*narrays, method=method, nan_policy='omit', axis=axis)
    xp_assert_close(res.data[~res.mask], xp.asarray(ref[~np.isnan(ref)]))
    xp_assert_close(res.mask, xp.asarray(np.isnan(ref)))


@make_xp_test_case(stats.obrientransform)
@skip_backend('jax.numpy', reason="JAX currently incompatible with marray")
@pytest.mark.parametrize('dtype', ['float32', 'float64'])
@pytest.mark.parametrize('n_arrays', [1, 3])
def test_obrientransform(dtype, n_arrays, xp):
    # obrientransform is not yet vectorized
    mxp, marrays, narrays = get_arrays(n_arrays, dtype=dtype, shape=(25,), xp=xp)
    res = stats.obrientransform(*marrays)
    ref = stats.obrientransform(*narrays, nan_policy='omit')
    for res_i, ref_i in zip(res, ref):
        xp_assert_close(res_i.data[~res_i.mask],
                        xp.asarray(ref_i[~np.isnan(ref_i)], dtype=getattr(xp, dtype)))



#
# Author: Travis Oliphant, March 2002
#
import warnings

import numpy as np
from numpy import (dot, diag, prod, logical_not, ravel, transpose,
                   conjugate, absolute, amax, sign, isfinite, triu)

from scipy._lib._util import _apply_over_batch, _deprecate_dtypes

# Local imports
from scipy.linalg import LinAlgError, LinAlgWarning
from ._misc import norm
from ._basic import solve, inv
from ._decomp_svd import svd
from ._decomp_schur import schur, rsf2csf
from ._expm_frechet import expm_frechet, expm_cond
from ._internal_matfuncs import recursive_schur_sqrtm, matrix_exponential
from ._linalg_pythran import _funm_loops  # type: ignore[import-not-found]

__all__ = ['expm', 'cosm', 'sinm', 'tanm', 'coshm', 'sinhm', 'tanhm', 'logm',
           'funm', 'signm', 'sqrtm', 'fractional_matrix_power', 'expm_frechet',
           'expm_cond', 'khatri_rao']

eps = np.finfo('d').eps
feps = np.finfo('f').eps

_array_precision = {'i': 1, 'l': 1, 'f': 0, 'd': 1, 'F': 0, 'D': 1}


###############################################################################
# Utility functions.


def _asarray_square(A):
    """
    Wraps asarray with the extra requirement that the input be a square matrix.

    The motivation is that the matfuncs module has real functions that have
    been lifted to square matrix functions.

    Parameters
    ----------
    A : array_like
        A square matrix.

    Returns
    -------
    out : ndarray
        An ndarray copy or view or other representation of A.

    """
    A = np.asarray(A)
    if len(A.shape) != 2 or A.shape[0] != A.shape[1]:
        raise ValueError('expected square array_like input')
    return A


def _maybe_real(A, B, tol=None):
    """
    Return either B or the real part of B, depending on properties of A and B.

    The motivation is that B has been computed as a complicated function of A,
    and B may be perturbed by negligible imaginary components.
    If A is real and B is complex with small imaginary components,
    then return a real copy of B.  The assumption in that case would be that
    the imaginary components of B are numerical artifacts.

    Parameters
    ----------
    A : ndarray
        Input array whose type is to be checked as real vs. complex.
    B : ndarray
        Array to be returned, possibly without its imaginary part.
    tol : float
        Absolute tolerance.

    Returns
    -------
    out : real or complex array
        Either the input array B or only the real part of the input array B.

    """
    # Note that booleans and integers compare as real.
    if np.isrealobj(A) and np.iscomplexobj(B):
        if tol is None:
            tol = {0: feps*1e3, 1: eps*1e6}[_array_precision[B.dtype.char]]
        if np.allclose(B.imag, 0.0, atol=tol):
            B = B.real
    return B


###############################################################################
# Matrix functions.


@_apply_over_batch(('A', 2))
def fractional_matrix_power(A, t):
    """
    Compute the fractional power of a matrix.

    Proceeds according to the discussion in section (6) of [1]_.

    Parameters
    ----------
    A : (N, N) array_like
        Matrix whose fractional power to evaluate.
    t : float
        Fractional power.

    Returns
    -------
    X : (N, N) array_like
        The fractional power of the matrix.

    References
    ----------
    .. [1] Nicholas J. Higham and Lijing Lin (2011)
           "A Schur-Pade Algorithm for Fractional Powers of a Matrix."
           SIAM Journal on Matrix Analysis and Applications,
           32 (3). pp. 1056-1078. ISSN 0895-4798.
           :doi:`10.1137/10081232X`

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.linalg import fractional_matrix_power
    >>> a = np.array([[1.0, 3.0], [1.0, 4.0]])
    >>> b = fractional_matrix_power(a, 0.5)
    >>> b
    array([[ 0.75592895,  1.13389342],
           [ 0.37796447,  1.88982237]])
    >>> np.dot(b, b)      # Verify square root
    array([[ 1.,  3.],
           [ 1.,  4.]])

    """
    # This fixes some issue with imports;
    # this function calls onenormest which is in scipy.sparse.
    A = _asarray_square(A)
    import scipy.linalg._matfuncs_inv_ssq
    return scipy.linalg._matfuncs_inv_ssq._fractional_matrix_power(A, t)


@_apply_over_batch(('A', 2))
def logm(A):
    """
    Compute matrix logarithm.

    The matrix logarithm is the inverse of
    expm: expm(logm(`A`)) == `A`

    Parameters
    ----------
    A : (N, N) array_like
        Matrix whose logarithm to evaluate

    Returns
    -------
    logm : (N, N) ndarray
        Matrix logarithm of `A`

    References
    ----------
    .. [1] Awad H. Al-Mohy and Nicholas J. Higham (2012)
           "Improved Inverse Scaling and Squaring Algorithms
           for the Matrix Logarithm."
           SIAM Journal on Scientific Computing, 34 (4). C152-C169.
           ISSN 1095-7197

    .. [2] Nicholas J. Higham (2008)
           "Functions of Matrices: Theory and Computation"
           ISBN 978-0-898716-46-7

    .. [3] Nicholas J. Higham and Lijing lin (2011)
           "A Schur-Pade Algorithm for Fractional Powers of a Matrix."
           SIAM Journal on Matrix Analysis and Applications,
           32 (3). pp. 1056-1078. ISSN 0895-4798

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.linalg import logm, expm
    >>> a = np.array([[1.0, 3.0], [1.0, 4.0]])
    >>> b = logm(a)
    >>> b
    array([[-1.02571087,  2.05142174],
           [ 0.68380725,  1.02571087]])
    >>> expm(b)         # Verify expm(logm(a)) returns a
    array([[ 1.,  3.],
           [ 1.,  4.]])

    """
    A = np.asarray(A)  # squareness checked in `_logm`
    # Avoid circular import ... this is OK, right?
    import scipy.linalg._matfuncs_inv_ssq
    F = scipy.linalg._matfuncs_inv_ssq._logm(A)
    F = _maybe_real(A, F)
    errtol = 1000*eps
    # TODO use a better error approximation
    with np.errstate(divide='ignore', invalid='ignore'):
        errest = norm(expm(F)-A, 1) / np.asarray(norm(A, 1), dtype=A.dtype).real[()]
    if not isfinite(errest) or errest >= errtol:
        message = f"logm result may be inaccurate, approximate err = {errest}"
        warnings.warn(message, RuntimeWarning, stacklevel=2)
    return F


def expm(A):
    """Compute the matrix exponential of an array.

    Array argument(s) of this function may have additional
    "batch" dimensions prepended to the core shape. In this case, the array is treated
    as a batch of lower-dimensional slices; see :ref:`linalg_batch` for details.

    Parameters
    ----------
    A : ndarray
        Input with last two dimensions are square ``(..., n, n)``.

    Returns
    -------
    eA : ndarray
        The resulting matrix exponential with the same shape of ``A``

    Notes
    -----
    Implements the algorithm given in [1]_, which is essentially a Pade
    approximation with a variable order that is decided based on the array
    data.

    For input with size ``n``, the memory usage is in the worst case in the
    order of ``8*(n**2)``. If the input data is not of single and double
    precision of real and complex dtypes, it is copied to a new array.

    For cases ``n >= 400``, the exact 1-norm computation cost, breaks even with
    1-norm estimation and from that point on the estimation scheme given in
    [2]_ is used to decide on the approximation order.

    References
    ----------
    .. [1] Awad H. Al-Mohy and Nicholas J. Higham, (2009), "A New Scaling
           and Squaring Algorithm for the Matrix Exponential", SIAM J. Matrix
           Anal. Appl. 31(3):970-989, :doi:`10.1137/09074721X`

    .. [2] Nicholas J. Higham and Francoise Tisseur (2000), "A Block Algorithm
           for Matrix 1-Norm Estimation, with an Application to 1-Norm
           Pseudospectra." SIAM J. Matrix Anal. Appl. 21(4):1185-1201,
           :doi:`10.1137/S0895479899356080`

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.linalg import expm, sinm, cosm

    Matrix version of the formula exp(0) = 1:

    >>> expm(np.zeros((3, 2, 2)))
    array([[[1., 0.],
            [0., 1.]],
    <BLANKLINE>
           [[1., 0.],
            [0., 1.]],
    <BLANKLINE>
           [[1., 0.],
            [0., 1.]]])

    Euler's identity (exp(i*theta) = cos(theta) + i*sin(theta))
    applied to a matrix:

    >>> a = np.array([[1.0, 2.0], [-1.0, 3.0]])
    >>> expm(1j*a)
    array([[ 0.42645930+1.89217551j, -2.13721484-0.97811252j],
           [ 1.06860742+0.48905626j, -1.71075555+0.91406299j]])
    >>> cosm(a) + 1j*sinm(a)
    array([[ 0.42645930+1.89217551j, -2.13721484-0.97811252j],
           [ 1.06860742+0.48905626j, -1.71075555+0.91406299j]])

    """
    a = np.asarray(A)
    _deprecate_dtypes("expm", a)

    if a.size == 1 and a.ndim < 2:
        return np.array([[np.exp(a.item())]])

    if a.ndim < 2:
        raise LinAlgError('The input array must be at least two-dimensional')
    if a.shape[-1] != a.shape[-2]:
        raise LinAlgError('Last 2 dimensions of the array must be square')

    # Empty array
    if min(*a.shape) == 0:
        dtype = expm(np.eye(2, dtype=a.dtype)).dtype
        return np.empty_like(a, dtype=dtype)

    # Scalar case
    if a.shape[-2:] == (1, 1):
        return np.exp(a)

    if not np.issubdtype(a.dtype, np.inexact):
        a = a.astype(np.float64)
    elif a.dtype == np.float16:
        a = a.astype(np.float32)

    # An explicit formula for 2x2 case exists (formula (2.2) in [1]_). However, without
    # Kahan's method, numerical instabilities can occur (See gh-19584). Hence removed
    # here until we have a more stable implementation.

    eA, info = matrix_exponential(a)
    if info != 0:
        if info <= -11:
            # We raise it from failed mallocs; negative LAPACK codes > -7
            raise MemoryError("scipy.linalg.expm could not allocate "
                            "sufficient memory while trying to compute the "
                            f"exponential (error code {info}).")
        else:
            # LAPACK wrong argument error or exact singularity.
            # Neither should happen.
            raise RuntimeError("scipy.linalg.expm: Internal LAPACK "
                                "error during the exponential computation "
                                f"(error code {info})")
    return eA


def sqrtm(A):
    """
    Compute, if exists, the matrix square root.

    The matrix square root of ``A`` is a matrix ``X`` such that ``X @ X = A``.
    Every square matrix is not guaranteed to have a matrix square root, for
    example, the array ``[[0, 1], [0, 0]]`` does not have a square root.

    Moreover, not every real matrix has a real square root. Hence, for
    real-valued matrices the return type can be complex if, numerically, there
    is an eigenvalue on the negative real axis.

    Array argument(s) of this function may have additional
    "batch" dimensions prepended to the core shape. In this case, the array is treated
    as a batch of lower-dimensional slices; see :ref:`linalg_batch` for details.

    Parameters
    ----------
    A : ndarray
        Input with last two dimensions are square ``(..., n, n)``.

    Returns
    -------
    sqrtm : ndarray
        Computed matrix squareroot of `A` with same size ``(..., n, n)``.

    Notes
    -----
    This function uses the Schur decomposition method to compute the matrix
    square root following [1]_ and for real matrices [2]_. Moreover, note
    that, there exist matrices that have square roots that are not polynomials
    in ``A``. For a classical example from [2]_, the matrix satisfies::

            [ a, a**2 + 1]**2     [-1,  0]
            [-1,       -a]     =  [ 0, -1]

    for any scalar ``a`` but it is not a polynomial in ``-I``. Thus, they will
    not be found by this function.

    References
    ----------
    .. [1] Edvin Deadman, Nicholas J. Higham, Rui Ralha (2013)
           "Blocked Schur Algorithms for Computing the Matrix Square Root,
           Lecture Notes in Computer Science, 7782. pp. 171-182.
           :doi:`10.1016/0024-3795(87)90118-2`
    .. [2] Nicholas J. Higham (1987) "Computing real square roots of a real
           matrix", Linear Algebra and its Applications, 88/89:405-430.
           :doi:`10.1016/0024-3795(87)90118-2`

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.linalg import sqrtm
    >>> a = np.array([[1.0, 3.0], [1.0, 4.0]])
    >>> r = sqrtm(a)
    >>> r
    array([[ 0.75592895,  1.13389342],
           [ 0.37796447,  1.88982237]])
    >>> r.dot(r)
    array([[ 1.,  3.],
           [ 1.,  4.]])

    """
    a = np.asarray(A)
    _deprecate_dtypes('sqrtm', a)
    if a.size == 1 and a.ndim < 2:
        return np.array([[np.exp(a.item())]])

    if a.ndim < 2:
        raise LinAlgError('The input array must be at least two-dimensional')
    if a.shape[-1] != a.shape[-2]:
        raise LinAlgError('Last 2 dimensions of the array must be square')

    # Empty array
    if min(*a.shape) == 0:
        dtype = sqrtm(np.eye(2, dtype=a.dtype)).dtype
        return np.empty_like(a, dtype=dtype)

    # Scalar case
    if a.shape[-2:] == (1, 1):
        return np.emath.sqrt(a)

    if not np.issubdtype(a.dtype, np.inexact):
        a = a.astype(np.float64)
    elif a.dtype == np.float16:
        a = a.astype(np.float32)
    elif a.dtype.char in 'G':
        a = a.astype(np.complex128)
    elif a.dtype.char in 'g':
        a = a.astype(np.float64)

    if a.dtype.char not in 'fdFD':
        raise TypeError("scipy.linalg.sqrtm is not supported for the data type"
                        f" {a.dtype}")

    res, isIllconditioned, isSingular, info = recursive_schur_sqrtm(a)
    if info < 0:
        raise LinAlgError(f"Internal error in scipy.linalg.sqrtm: {info}")

    if isSingular or isIllconditioned:
        if isSingular:
            msg = ("Matrix is singular. The result might be inaccurate or the"
                   " array might not have a square root.")
        else:
            msg = ("Matrix is ill-conditioned. The result might be inaccurate"
                   " or the array might not have a square root.")
        warnings.warn(msg, LinAlgWarning, stacklevel=2)

    return res


@_apply_over_batch(('A', 2))
def cosm(A):
    """
    Compute the matrix cosine.

    This routine uses expm to compute the matrix exponentials.

    Parameters
    ----------
    A : (N, N) array_like
        Input array

    Returns
    -------
    cosm : (N, N) ndarray
        Matrix cosine of A

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.linalg import expm, sinm, cosm

    Euler's identity (exp(i*theta) = cos(theta) + i*sin(theta))
    applied to a matrix:

    >>> a = np.array([[1.0, 2.0], [-1.0, 3.0]])
    >>> expm(1j*a)
    array([[ 0.42645930+1.89217551j, -2.13721484-0.97811252j],
           [ 1.06860742+0.48905626j, -1.71075555+0.91406299j]])
    >>> cosm(a) + 1j*sinm(a)
    array([[ 0.42645930+1.89217551j, -2.13721484-0.97811252j],
           [ 1.06860742+0.48905626j, -1.71075555+0.91406299j]])

    """
    A = _asarray_square(A)
    if np.iscomplexobj(A):
        return 0.5*(expm(1j*A) + expm(-1j*A))
    else:
        return expm(1j*A).real


@_apply_over_batch(('A', 2))
def sinm(A):
    """
    Compute the matrix sine.

    This routine uses expm to compute the matrix exponentials.

    Parameters
    ----------
    A : (N, N) array_like
        Input array.

    Returns
    -------
    sinm : (N, N) ndarray
        Matrix sine of `A`

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.linalg import expm, sinm, cosm

    Euler's identity (exp(i*theta) = cos(theta) + i*sin(theta))
    applied to a matrix:

    >>> a = np.array([[1.0, 2.0], [-1.0, 3.0]])
    >>> expm(1j*a)
    array([[ 0.42645930+1.89217551j, -2.13721484-0.97811252j],
           [ 1.06860742+0.48905626j, -1.71075555+0.91406299j]])
    >>> cosm(a) + 1j*sinm(a)
    array([[ 0.42645930+1.89217551j, -2.13721484-0.97811252j],
           [ 1.06860742+0.48905626j, -1.71075555+0.91406299j]])

    """
    A = _asarray_square(A)
    if np.iscomplexobj(A):
        return -0.5j*(expm(1j*A) - expm(-1j*A))
    else:
        return expm(1j*A).imag


@_apply_over_batch(('A', 2))
def tanm(A):
    """
    Compute the matrix tangent.

    This routine uses expm to compute the matrix exponentials.

    Parameters
    ----------
    A : (N, N) array_like
        Input array.

    Returns
    -------
    tanm : (N, N) ndarray
        Matrix tangent of `A`

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.linalg import tanm, sinm, cosm
    >>> a = np.array([[1.0, 3.0], [1.0, 4.0]])
    >>> t = tanm(a)
    >>> t
    array([[ -2.00876993,  -8.41880636],
           [ -2.80626879, -10.42757629]])

    Verify tanm(a) = sinm(a).dot(inv(cosm(a)))

    >>> s = sinm(a)
    >>> c = cosm(a)
    >>> s.dot(np.linalg.inv(c))
    array([[ -2.00876993,  -8.41880636],
           [ -2.80626879, -10.42757629]])

    """
    A = _asarray_square(A)
    return _maybe_real(A, solve(cosm(A), sinm(A)))


@_apply_over_batch(('A', 2))
def coshm(A):
    """
    Compute the hyperbolic matrix cosine.

    This routine uses expm to compute the matrix exponentials.

    Parameters
    ----------
    A : (N, N) array_like
        Input array.

    Returns
    -------
    coshm : (N, N) ndarray
        Hyperbolic matrix cosine of `A`

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.linalg import tanhm, sinhm, coshm
    >>> a = np.array([[1.0, 3.0], [1.0, 4.0]])
    >>> c = coshm(a)
    >>> c
    array([[ 11.24592233,  38.76236492],
           [ 12.92078831,  50.00828725]])

    Verify tanhm(a) = sinhm(a).dot(inv(coshm(a)))

    >>> t = tanhm(a)
    >>> s = sinhm(a)
    >>> t - s.dot(np.linalg.inv(c))
    array([[  2.72004641e-15,   4.55191440e-15],
           [  0.00000000e+00,  -5.55111512e-16]])

    """
    A = _asarray_square(A)
    return _maybe_real(A, 0.5 * (expm(A) + expm(-A)))


@_apply_over_batch(('A', 2))
def sinhm(A):
    """
    Compute the hyperbolic matrix sine.

    This routine uses expm to compute the matrix exponentials.

    Parameters
    ----------
    A : (N, N) array_like
        Input array.

    Returns
    -------
    sinhm : (N, N) ndarray
        Hyperbolic matrix sine of `A`

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.linalg import tanhm, sinhm, coshm
    >>> a = np.array([[1.0, 3.0], [1.0, 4.0]])
    >>> s = sinhm(a)
    >>> s
    array([[ 10.57300653,  39.28826594],
           [ 13.09608865,  49.86127247]])

    Verify tanhm(a) = sinhm(a).dot(inv(coshm(a)))

    >>> t = tanhm(a)
    >>> c = coshm(a)
    >>> t - s.dot(np.linalg.inv(c))
    array([[  2.72004641e-15,   4.55191440e-15],
           [  0.00000000e+00,  -5.55111512e-16]])

    """
    A = _asarray_square(A)
    return _maybe_real(A, 0.5 * (expm(A) - expm(-A)))


@_apply_over_batch(('A', 2))
def tanhm(A):
    """
    Compute the hyperbolic matrix tangent.

    This routine uses expm to compute the matrix exponentials.

    Parameters
    ----------
    A : (N, N) array_like
        Input array

    Returns
    -------
    tanhm : (N, N) ndarray
        Hyperbolic matrix tangent of `A`

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.linalg import tanhm, sinhm, coshm
    >>> a = np.array([[1.0, 3.0], [1.0, 4.0]])
    >>> t = tanhm(a)
    >>> t
    array([[ 0.3428582 ,  0.51987926],
           [ 0.17329309,  0.86273746]])

    Verify tanhm(a) = sinhm(a).dot(inv(coshm(a)))

    >>> s = sinhm(a)
    >>> c = coshm(a)
    >>> t - s.dot(np.linalg.inv(c))
    array([[  2.72004641e-15,   4.55191440e-15],
           [  0.00000000e+00,  -5.55111512e-16]])

    """
    A = _asarray_square(A)
    return _maybe_real(A, solve(coshm(A), sinhm(A)))


@_apply_over_batch(('A', 2))
def funm(A, func, disp=True):
    """
    Evaluate a matrix function specified by a callable.

    Returns the value of matrix-valued function ``f`` at `A`. The
    function ``f`` is an extension of the scalar-valued function `func`
    to matrices.

    Parameters
    ----------
    A : (N, N) array_like
        Matrix at which to evaluate the function
    func : callable
        Callable object that evaluates a scalar function f.
        Must be vectorized (eg. using vectorize).
    disp : bool, optional
        Print warning if error in the result is estimated large
        instead of returning estimated error. (Default: True)

    Returns
    -------
    funm : (N, N) ndarray
        Value of the matrix function specified by func evaluated at `A`
    errest : float
        (if disp == False)

        1-norm of the estimated error, ||err||_1 / ||A||_1

    Notes
    -----
    This function implements the general algorithm based on Schur decomposition
    (Algorithm 9.1.1. in [1]_).

    If the input matrix is known to be diagonalizable, then relying on the
    eigendecomposition is likely to be faster. For example, if your matrix is
    Hermitian, you can do

    >>> from scipy.linalg import eigh
    >>> def funm_herm(a, func, check_finite=False):
    ...     w, v = eigh(a, check_finite=check_finite)
    ...     ## if you further know that your matrix is positive semidefinite,
    ...     ## you can optionally guard against precision errors by doing
    ...     # w = np.maximum(w, 0)
    ...     w = func(w)
    ...     return (v * w).dot(v.conj().T)

    References
    ----------
    .. [1] Gene H. Golub, Charles F. van Loan, Matrix Computations 4th ed.

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.linalg import funm
    >>> a = np.array([[1.0, 3.0], [1.0, 4.0]])
    >>> funm(a, lambda x: x*x)
    array([[  4.,  15.],
           [  5.,  19.]])
    >>> a.dot(a)
    array([[  4.,  15.],
           [  5.,  19.]])

    """
    A = _asarray_square(A)
    # Perform Shur decomposition (lapack ?gees)
    T, Z = schur(A)
    T, Z = rsf2csf(T, Z)
    n, n = T.shape
    F = diag(func(diag(T)))  # apply function to diagonal elements
    F = F.astype(T.dtype.char)  # e.g., when F is real but T is complex

    minden = abs(T[0, 0])

    # implement Algorithm 11.1.1 from Golub and Van Loan
    #                 "matrix Computations."
    F, minden = _funm_loops(F, T, n, minden)

    F = dot(dot(Z, F), transpose(conjugate(Z)))
    F = _maybe_real(A, F)

    tol = {0: feps, 1: eps}[_array_precision[F.dtype.char]]
    if minden == 0.0:
        minden = tol
    err = min(1, max(tol, (tol/minden)*norm(triu(T, 1), 1)))
    if prod(ravel(logical_not(isfinite(F))), axis=0):
        err = np.inf
    if disp:
        if err > 1000*tol:
            print("funm result may be inaccurate, approximate err =", err)
        return F
    else:
        return F, err


@_apply_over_batch(('A', 2))
def signm(A):
    """
    Matrix sign function.

    Extension of the scalar sign(x) to matrices.

    Parameters
    ----------
    A : (N, N) array_like
        Matrix at which to evaluate the sign function

    Returns
    -------
    signm : (N, N) ndarray
        Value of the sign function at `A`

    Examples
    --------
    >>> from scipy.linalg import signm, eigvals
    >>> a = [[1,2,3], [1,2,1], [1,1,1]]
    >>> eigvals(a)
    array([ 4.12488542+0.j, -0.76155718+0.j,  0.63667176+0.j])
    >>> eigvals(signm(a))
    array([-1.+0.j,  1.+0.j,  1.+0.j])

    """
    A = _asarray_square(A)

    def rounded_sign(x):
        rx = np.real(x)
        if rx.dtype.char == 'f':
            c = 1e3*feps*amax(x)
        else:
            c = 1e3*eps*amax(x)
        return sign((absolute(rx) > c) * rx)
    result, errest = funm(A, rounded_sign, disp=0)
    errtol = {0: 1e3*feps, 1: 1e3*eps}[_array_precision[result.dtype.char]]
    if errest < errtol:
        return result

    # Handle signm of defective matrices:

    # See "E.D.Denman and J.Leyva-Ramos, Appl.Math.Comp.,
    # 8:237-250,1981" for how to improve the following (currently a
    # rather naive) iteration process:

    # a = result # sometimes iteration converges faster but where??

    # Shifting to avoid zero eigenvalues. How to ensure that shifting does
    # not change the spectrum too much?
    vals = svd(A, compute_uv=False)
    max_sv = np.amax(vals)
    # min_nonzero_sv = vals[(vals>max_sv*errtol).tolist().count(1)-1]
    # c = 0.5/min_nonzero_sv
    c = 0.5/max_sv
    S0 = A + c*np.identity(A.shape[0])
    prev_errest = errest
    for i in range(100):
        iS0 = inv(S0)
        S0 = 0.5*(S0 + iS0)
        Pp = 0.5*(dot(S0, S0)+S0)
        errest = norm(dot(Pp, Pp)-Pp, 1)
        if errest < errtol or prev_errest == errest:
            break
        prev_errest = errest
    if not isfinite(errest) or errest >= errtol:
        print("signm result may be inaccurate, approximate err =", errest)
    return S0


@_apply_over_batch(('a', 2), ('b', 2))
def khatri_rao(a, b):
    r"""
    Khatri-Rao product of two matrices.

    A column-wise Kronecker product of two matrices

    Parameters
    ----------
    a : (n, k) array_like
        Input array
    b : (m, k) array_like
        Input array

    Returns
    -------
    c:  (n*m, k) ndarray
        Khatri-rao product of `a` and `b`.

    Notes
    -----
    The mathematical definition of the Khatri-Rao product is:

    .. math::

        (A_{ij}  \bigotimes B_{ij})_{ij}

    which is the Kronecker product of every column of A and B, e.g.::

        c = np.vstack([np.kron(a[:, k], b[:, k]) for k in range(b.shape[1])]).T

    Examples
    --------
    >>> import numpy as np
    >>> from scipy import linalg
    >>> a = np.array([[1, 2, 3], [4, 5, 6]])
    >>> b = np.array([[3, 4, 5], [6, 7, 8], [2, 3, 9]])
    >>> linalg.khatri_rao(a, b)
    array([[ 3,  8, 15],
           [ 6, 14, 24],
           [ 2,  6, 27],
           [12, 20, 30],
           [24, 35, 48],
           [ 8, 15, 54]])

    """
    a = np.asarray(a)
    b = np.asarray(b)

    if not (a.ndim == 2 and b.ndim == 2):
        raise ValueError("The both arrays should be 2-dimensional.")

    if not a.shape[1] == b.shape[1]:
        raise ValueError("The number of columns for both arrays "
                         "should be equal.")

    # accommodate empty arrays
    if a.size == 0 or b.size == 0:
        m = a.shape[0] * b.shape[0]
        n = a.shape[1]
        return np.empty_like(a, shape=(m, n))

    # c = np.vstack([np.kron(a[:, k], b[:, k]) for k in range(b.shape[1])]).T
    c = a[..., :, np.newaxis, :] * b[..., np.newaxis, :, :]
    return c.reshape((-1,) + c.shape[2:])



import sys
import threading
import warnings

import numpy as np
from numpy import array, finfo, arange, eye, all, unique, ones, dot
from numpy.exceptions import ComplexWarning
from numpy.testing import (
        assert_array_almost_equal, assert_almost_equal,
        assert_equal, assert_array_equal, assert_, assert_allclose)
import pytest
from pytest import raises as assert_raises, warns as assert_warns

import scipy.linalg
from scipy.linalg import norm, inv
from scipy.sparse import (issparse, SparseEfficiencyWarning,
                          csc_array, csr_array, coo_array, dia_array,
                          dok_array, lil_array, bsr_array,
                          eye_array, random_array, tril, triu, kron)
from scipy.sparse.linalg import SuperLU
from scipy.sparse.linalg._dsolve import (spsolve, use_solver, splu, spilu,
        MatrixRankWarning, _superlu, spsolve_triangular, factorized,
        is_sptriangular, spbandwidth)
import scipy.sparse

from scipy._lib._testutils import check_free_memory


# scikits.umfpack is not a SciPy dependency but it is optionally used in
# dsolve, so check whether it's available
try:
    import scikits.umfpack as umfpack
    has_umfpack = True
except ImportError:
    has_umfpack = False

def toarray(a):
    if issparse(a):
        return a.toarray()
    else:
        return a


def setup_bug_8278():
    N = 2 ** 6
    h = 1/N
    Ah1D = dia_array(([-1, 2, -1], [-1, 0, 1]), shape=(N-1, N-1))/(h**2)
    eyeN = eye_array(N - 1)
    A = (kron(eyeN, kron(eyeN, Ah1D))
         + kron(eyeN, kron(Ah1D, eyeN))
         + kron(Ah1D, kron(eyeN, eyeN)))
    b = np.random.rand((N-1)**3)
    return A, b


class TestFactorized:
    def setup_method(self):
        n = 5
        d = arange(n) + 1
        self.n = n
        self.A = dia_array(((d, 2*d, d[::-1]), (-3, 0, 5)), shape=(n,n)).tocsc()

    def _check_singular(self):
        A = csc_array((5,5), dtype='d')
        b = ones(5)
        assert_array_almost_equal(0. * b, factorized(A)(b))

    def _check_non_singular(self):
        # Make a diagonal dominant, to make sure it is not singular
        n = 5
        rng = np.random.default_rng(14332)
        a = csc_array(rng.random((n, n)))
        b = ones(n)

        expected = splu(a).solve(b)
        assert_array_almost_equal(factorized(a)(b), expected)

    def test_singular_without_umfpack(self):
        use_solver(useUmfpack=False)
        with assert_raises(RuntimeError, match="Factor is exactly singular"):
            self._check_singular()

    @pytest.mark.skipif(not has_umfpack, reason="umfpack not available")
    def test_singular_with_umfpack(self):
        use_solver(useUmfpack=True)
        with warnings.catch_warnings():
            msg = "divide by zero encountered in double_scalars"
            warnings.filterwarnings("ignore", msg, RuntimeWarning)
            assert_warns(umfpack.UmfpackWarning, self._check_singular)

    def test_non_singular_without_umfpack(self):
        use_solver(useUmfpack=False)
        self._check_non_singular()

    @pytest.mark.skipif(not has_umfpack, reason="umfpack not available")
    def test_non_singular_with_umfpack(self):
        use_solver(useUmfpack=True)
        self._check_non_singular()

    def test_cannot_factorize_nonsquare_matrix_without_umfpack(self):
        use_solver(useUmfpack=False)
        msg = "can only factor square matrices"
        with assert_raises(ValueError, match=msg):
            factorized(self.A[:, :4])

    @pytest.mark.skipif(not has_umfpack, reason="umfpack not available")
    def test_factorizes_nonsquare_matrix_with_umfpack(self):
        use_solver(useUmfpack=True)
        # does not raise
        factorized(self.A[:,:4])

    def test_call_with_incorrectly_sized_matrix_without_umfpack(self):
        use_solver(useUmfpack=False)
        solve = factorized(self.A)

        rng = np.random.default_rng(230498)
        b = rng.random(4)
        B = rng.random((4, 3))
        BB = rng.random((self.n, 3, 9))

        with assert_raises(ValueError, match="is of incompatible size"):
            solve(b)
        with assert_raises(ValueError, match="is of incompatible size"):
            solve(B)
        with assert_raises(ValueError,
                           match="object too deep for desired array"):
            solve(BB)

    @pytest.mark.skipif(not has_umfpack, reason="umfpack not available")
    def test_call_with_incorrectly_sized_matrix_with_umfpack(self):
        use_solver(useUmfpack=True)
        solve = factorized(self.A)

        rng = np.random.default_rng(643095823)
        b = rng.random(4)
        B = rng.random((4, 3))
        BB = rng.random((self.n, 3, 9))

        # does not raise
        solve(b)
        msg = "object too deep for desired array"
        with assert_raises(ValueError, match=msg):
            solve(B)
        with assert_raises(ValueError, match=msg):
            solve(BB)

    def test_call_with_cast_to_complex_without_umfpack(self):
        use_solver(useUmfpack=False)
        solve = factorized(self.A)
        rng = np.random.default_rng(23454)
        b = rng.random(4)
        for t in [np.complex64, np.complex128]:
            with assert_raises(TypeError, match="Cannot cast array data"):
                solve(b.astype(t))

    @pytest.mark.skipif(not has_umfpack, reason="umfpack not available")
    def test_call_with_cast_to_complex_with_umfpack(self):
        use_solver(useUmfpack=True)
        solve = factorized(self.A)
        rng = np.random.default_rng(23454)
        b = rng.random(4)
        for t in [np.complex64, np.complex128]:
            assert_warns(ComplexWarning, solve, b.astype(t))

    @pytest.mark.skipif(not has_umfpack, reason="umfpack not available")
    def test_assume_sorted_indices_flag(self):
        # a sparse matrix with unsorted indices
        unsorted_inds = np.array([2, 0, 1, 0])
        data = np.array([10, 16, 5, 0.4])
        indptr = np.array([0, 1, 2, 4])
        A = csc_array((data, unsorted_inds, indptr), (3, 3))
        b = ones(3)

        # should raise when incorrectly assuming indices are sorted
        use_solver(useUmfpack=True, assumeSortedIndices=True)
        with assert_raises(RuntimeError,
                           match="UMFPACK_ERROR_invalid_matrix"):
            factorized(A)

        # should sort indices and succeed when not assuming indices are sorted
        use_solver(useUmfpack=True, assumeSortedIndices=False)
        expected = splu(A.copy()).solve(b)

        assert_equal(A.has_sorted_indices, 0)
        assert_array_almost_equal(factorized(A)(b), expected)

    @pytest.mark.slow
    @pytest.mark.skipif(not has_umfpack, reason="umfpack not available")
    def test_bug_8278(self):
        check_free_memory(8000)
        use_solver(useUmfpack=True)
        A, b = setup_bug_8278()
        A = A.tocsc()
        f = factorized(A)
        x = f(b)
        assert_array_almost_equal(A @ x, b)


class TestLinsolve:
    def setup_method(self):
        use_solver(useUmfpack=False)

    def test_singular(self):
        A = csc_array((5,5), dtype='d')
        b = array([1, 2, 3, 4, 5],dtype='d')
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", "Matrix is exactly singular", MatrixRankWarning)
            x = spsolve(A, b)
        assert_(not np.isfinite(x).any())

    def test_singular_gh_3312(self):
        # "Bad" test case that leads SuperLU to call LAPACK with invalid
        # arguments. Check that it fails moderately gracefully.
        ij = np.array([(17, 0), (17, 6), (17, 12), (10, 13)], dtype=np.int32)
        v = np.array([0.284213, 0.94933781, 0.15767017, 0.38797296])
        A = csc_array((v, ij.T), shape=(20, 20))
        b = np.arange(20)

        try:
            # should either raise a runtime error or return value
            # appropriate for singular input (which yields the warning)
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore", "Matrix is exactly singular", MatrixRankWarning)
                x = spsolve(A, b)
            assert not np.isfinite(x).any()
        except RuntimeError:
            pass

    @pytest.mark.parametrize('format', ['csc', 'csr'])
    @pytest.mark.parametrize('idx_dtype', [np.int32, np.int64])
    def test_twodiags(self, format: str, idx_dtype: np.dtype):
        A = dia_array(([[1, 2, 3, 4, 5], [6, 5, 8, 9, 10]], [0, 1]),
                        shape=(5, 5)).asformat(format)
        b = array([1, 2, 3, 4, 5])

        # condition number of A
        cond_A = norm(A.toarray(), 2) * norm(inv(A.toarray()), 2)

        for t in ['f','d','F','D']:
            eps = finfo(t).eps  # floating point epsilon
            b = b.astype(t)
            Asp = A.astype(t)
            Asp.indices = Asp.indices.astype(idx_dtype, copy=False)
            Asp.indptr = Asp.indptr.astype(idx_dtype, copy=False)

            x = spsolve(Asp, b)
            assert_(norm(b - Asp@x) < 10 * cond_A * eps)

    def test_bvector_smoketest(self):
        Adense = array([[0., 1., 1.],
                        [1., 0., 1.],
                        [0., 0., 1.]])
        As = csc_array(Adense)
        rng = np.random.default_rng(1234)
        x = rng.standard_normal(3)
        b = As@x
        x2 = spsolve(As, b)

        assert_array_almost_equal(x, x2)

    def test_bmatrix_smoketest(self):
        Adense = array([[0., 1., 1.],
                        [1., 0., 1.],
                        [0., 0., 1.]])
        As = csc_array(Adense)
        rng = np.random.default_rng(1234)
        x = rng.standard_normal((3, 4))
        Bdense = As.dot(x)
        Bs = csc_array(Bdense)
        x2 = spsolve(As, Bs)
        assert_array_almost_equal(x, x2.toarray())

    def test_non_square(self):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', SparseEfficiencyWarning)
            # A is not square.
            A = ones((3, 4))
            b = ones((4, 1))
            assert_raises(ValueError, spsolve, A, b)
            # A2 and b2 have incompatible shapes.
            A2 = csc_array(eye(3))
            b2 = array([1.0, 2.0])
            assert_raises(ValueError, spsolve, A2, b2)

    def test_example_comparison(self):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', SparseEfficiencyWarning)
            row = array([0,0,1,2,2,2])
            col = array([0,2,2,0,1,2])
            data = array([1,2,3,-4,5,6])
            sM = csr_array((data,(row,col)), shape=(3,3), dtype=float)
            M = sM.toarray()

            row = array([0,0,1,1,0,0])
            col = array([0,2,1,1,0,0])
            data = array([1,1,1,1,1,1])
            sN = csr_array((data, (row,col)), shape=(3,3), dtype=float)
            N = sN.toarray()

            sX = spsolve(sM, sN)
            X = scipy.linalg.solve(M, N)

            assert_array_almost_equal(X, sX.toarray())

    @pytest.mark.skipif(not has_umfpack, reason="umfpack not available")
    def test_shape_compatibility(self):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', SparseEfficiencyWarning)
            use_solver(useUmfpack=True)
            A = csc_array([[1., 0], [0, 2]])
            bs = [
                [1, 6],
                array([1, 6]),
                [[1], [6]],
                array([[1], [6]]),
                csc_array([[1], [6]]),
                csr_array([[1], [6]]),
                dok_array([[1], [6]]),
                bsr_array([[1], [6]]),
                array([[1., 2., 3.], [6., 8., 10.]]),
                csc_array([[1., 2., 3.], [6., 8., 10.]]),
                csr_array([[1., 2., 3.], [6., 8., 10.]]),
                dok_array([[1., 2., 3.], [6., 8., 10.]]),
                bsr_array([[1., 2., 3.], [6., 8., 10.]]),
            ]

        for b in bs:
            x = np.linalg.solve(A.toarray(), toarray(b))
            for spmattype in [csc_array, csr_array, dok_array, lil_array]:
                x1 = spsolve(spmattype(A), b, use_umfpack=True)
                x2 = spsolve(spmattype(A), b, use_umfpack=False)

                # check solution
                if x.ndim == 2 and x.shape[1] == 1:
                    # interprets also these as "vectors"
                    x = x.ravel()

                assert_array_almost_equal(toarray(x1), x,
                                          err_msg=repr((b, spmattype, 1)))
                assert_array_almost_equal(toarray(x2), x,
                                          err_msg=repr((b, spmattype, 2)))

                # dense vs. sparse output  ("vectors" are always dense)
                if issparse(b) and x.ndim > 1:
                    assert_(issparse(x1), repr((b, spmattype, 1)))
                    assert_(issparse(x2), repr((b, spmattype, 2)))
                else:
                    assert_(isinstance(x1, np.ndarray), repr((b, spmattype, 1)))
                    assert_(isinstance(x2, np.ndarray), repr((b, spmattype, 2)))

                # check output shape
                if x.ndim == 1:
                    # "vector"
                    assert_equal(x1.shape, (A.shape[1],))
                    assert_equal(x2.shape, (A.shape[1],))
                else:
                    # "matrix"
                    assert_equal(x1.shape, x.shape)
                    assert_equal(x2.shape, x.shape)

        A = csc_array((3, 3))
        b = csc_array((1, 3))
        assert_raises(ValueError, spsolve, A, b)

    def test_ndarray_support(self):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', SparseEfficiencyWarning)
            A = array([[1., 2.], [2., 0.]])
            x = array([[1., 1.], [0.5, -0.5]])
            b = array([[2., 0.], [2., 2.]])

            assert_array_almost_equal(x, spsolve(A, b))

    def test_gssv_badinput(self):
        N = 10
        d = arange(N) + 1.0
        A = dia_array(((d, 2*d, d[::-1]), (-3, 0, 5)), shape=(N, N))

        for container in (csc_array, csr_array):
            A = container(A)
            b = np.arange(N)

            def not_c_contig(x):
                return x.repeat(2)[::2]

            def not_1dim(x):
                return x[:,None]

            def bad_type(x):
                return x.astype(bool)

            def too_short(x):
                return x[:-1]

            badops = [not_c_contig, not_1dim, bad_type, too_short]

            for badop in badops:
                msg = f"{container!r} {badop!r}"
                # Not C-contiguous
                assert_raises((ValueError, TypeError), _superlu.gssv,
                              N, A.nnz, badop(A.data), A.indices, A.indptr,
                              b, int(A.format == 'csc'), err_msg=msg)
                assert_raises((ValueError, TypeError), _superlu.gssv,
                              N, A.nnz, A.data, badop(A.indices), A.indptr,
                              b, int(A.format == 'csc'), err_msg=msg)
                assert_raises((ValueError, TypeError), _superlu.gssv,
                              N, A.nnz, A.data, A.indices, badop(A.indptr),
                              b, int(A.format == 'csc'), err_msg=msg)

    def test_sparsity_preservation(self):
        ident = csc_array([
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 1]])
        b = csc_array([
            [0, 1],
            [1, 0],
            [0, 0]])
        x = spsolve(ident, b)
        assert_equal(ident.nnz, 3)
        assert_equal(b.nnz, 2)
        assert_equal(x.nnz, 2)
        assert_allclose(x.toarray(), b.toarray(), atol=1e-12, rtol=1e-12)

    def test_dtype_cast(self):
        A_real = csr_array([[1, 2, 0],
                            [0, 0, 3],
                            [4, 0, 5]])
        A_complex = csr_array([[1, 2, 0],
                               [0, 0, 3],
                               [4, 0, 5 + 1j]])
        b_real = np.array([1,1,1])
        b_complex = np.array([1,1,1]) + 1j*np.array([1,1,1])
        x = spsolve(A_real, b_real)
        assert_(np.issubdtype(x.dtype, np.floating))
        x = spsolve(A_real, b_complex)
        assert_(np.issubdtype(x.dtype, np.complexfloating))
        x = spsolve(A_complex, b_real)
        assert_(np.issubdtype(x.dtype, np.complexfloating))
        x = spsolve(A_complex, b_complex)
        assert_(np.issubdtype(x.dtype, np.complexfloating))

    @pytest.mark.slow
    @pytest.mark.skipif(not has_umfpack, reason="umfpack not available")
    def test_bug_8278(self):
        check_free_memory(8000)
        use_solver(useUmfpack=True)
        A, b = setup_bug_8278()
        x = spsolve(A, b)
        assert_array_almost_equal(A @ x, b)


class TestSplu:
    def setup_method(self):
        use_solver(useUmfpack=False)
        n = 40
        d = arange(n) + 1
        self.n = n
        self.A = dia_array(((d, 2*d, d[::-1]), (-3, 0, 5)), shape=(n, n)).tocsc()

    def _smoketest(self, spxlu, check, dtype, idx_dtype):
        if np.issubdtype(dtype, np.complexfloating):
            A = self.A + 1j*self.A.T
        else:
            A = self.A

        A = A.astype(dtype)
        A.indices = A.indices.astype(idx_dtype, copy=False)
        A.indptr = A.indptr.astype(idx_dtype, copy=False)
        lu = spxlu(A)

        rng = np.random.RandomState(1234)

        # Input shapes
        for k in [None, 1, 2, self.n, self.n+2]:
            msg = f"k={k!r}"

            if k is None:
                b = rng.rand(self.n)
            else:
                b = rng.rand(self.n, k)

            if np.issubdtype(dtype, np.complexfloating):
                b = b + 1j*rng.rand(*b.shape)
            b = b.astype(dtype)

            x = lu.solve(b)
            check(A, b, x, msg)

            x = lu.solve(b, 'T')
            check(A.T, b, x, msg)

            x = lu.solve(b, 'H')
            check(A.T.conj(), b, x, msg)

    def test_splu_smoketest(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", SparseEfficiencyWarning)
            self._internal_test_splu_smoketest()

    def _internal_test_splu_smoketest(self):
        # Check that splu works at all
        def check(A, b, x, msg=""):
            eps = np.finfo(A.dtype).eps
            r = A @ x
            assert_(abs(r - b).max() < 1e3*eps, msg)

        for dtype in [np.float32, np.float64, np.complex64, np.complex128]:
            for idx_dtype in [np.int32, np.int64]:
                self._smoketest(splu, check, dtype, idx_dtype)

    def test_spilu_smoketest(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", SparseEfficiencyWarning)
            self._internal_test_spilu_smoketest()

    def _internal_test_spilu_smoketest(self):
        errors = []

        def check(A, b, x, msg=""):
            r = A @ x
            err = abs(r - b).max()
            assert_(err < 1e-2, msg)
            if b.dtype in (np.float64, np.complex128):
                errors.append(err)

        for dtype in [np.float32, np.float64, np.complex64, np.complex128]:
            for idx_dtype in [np.int32, np.int64]:
                self._smoketest(spilu, check, dtype, idx_dtype)

        assert_(max(errors) > 1e-5)

    def test_spilu_drop_rule(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", SparseEfficiencyWarning)
            # Test passing in the drop_rule argument to spilu.
            A = eye_array(2)

            rules = [
                b'basic,area'.decode('ascii'),  # unicode
                b'basic,area',  # ascii
                [b'basic', b'area'.decode('ascii')]
            ]
            for rule in rules:
                # Argument should be accepted
                assert_(isinstance(spilu(A, drop_rule=rule), SuperLU))

    def test_splu_nnz0(self):
        A = csc_array((5,5), dtype='d')
        assert_raises(RuntimeError, splu, A)

    def test_spilu_nnz0(self):
        A = csc_array((5,5), dtype='d')
        assert_raises(RuntimeError, spilu, A)

    def test_splu_basic(self):
        # Test basic splu functionality.
        n = 30
        rng = np.random.RandomState(12)
        a = rng.rand(n, n)
        a[a < 0.95] = 0
        # First test with a singular matrix
        a[:, 0] = 0
        a_ = csc_array(a)
        # Matrix is exactly singular
        assert_raises(RuntimeError, splu, a_)

        # Make a diagonal dominant, to make sure it is not singular
        a += 4*eye(n)
        a_ = csc_array(a)
        lu = splu(a_)
        b = ones(n)
        x = lu.solve(b)
        assert_almost_equal(dot(a, x), b)

    def test_splu_perm(self):
        # Test the permutation vectors exposed by splu.
        n = 30
        rng = np.random.default_rng(1342354)
        a = rng.random((n, n))
        a[a < 0.95] = 0
        # Make a diagonal dominant, to make sure it is not singular
        a += 4*eye(n)
        a_ = csc_array(a)
        lu = splu(a_)
        # Check that the permutation indices do belong to [0, n-1].
        for perm in (lu.perm_r, lu.perm_c):
            assert_(all(perm > -1))
            assert_(all(perm < n))
            assert_equal(len(unique(perm)), len(perm))

        # Now make a symmetric, and test that the two permutation vectors are
        # the same
        # Note: a += a.T relies on undefined behavior.
        a = a + a.T
        a_ = csc_array(a)
        lu = splu(a_)
        assert_array_equal(lu.perm_r, lu.perm_c)

    @pytest.mark.parametrize("splu_fun, rtol", [(splu, 1e-7), (spilu, 1e-1)])
    def test_natural_permc(self, splu_fun, rtol):
        # Test that the "NATURAL" permc_spec does not permute the matrix
        rng = np.random.RandomState(42)
        n = 500
        p = 0.01
        A = random_array((n, n), density=p, random_state=rng)
        x = rng.rand(n)
        # Make A diagonal dominant to make sure it is not singular
        A += (n + 1) * eye_array(n)
        A_ = csc_array(A)
        b = A_ @ x

        # without permc_spec, permutation is not identity
        lu = splu_fun(A_)
        assert_(np.any(lu.perm_c != np.arange(n)))

        # with permc_spec="NATURAL", permutation is identity
        lu = splu_fun(A_, permc_spec="NATURAL")
        assert_array_equal(lu.perm_c, np.arange(n))

        # Also, lu decomposition is valid
        x2 = lu.solve(b)
        assert_allclose(x, x2, rtol=rtol)

    @pytest.mark.skipif(not hasattr(sys, 'getrefcount'), reason="no sys.getrefcount")
    def test_lu_refcount(self):
        # Test that we are keeping track of the reference count with splu.
        n = 30
        rng = np.random.default_rng(1342354)
        a = rng.random((n, n))
        a[a < 0.95] = 0
        # Make a diagonal dominant, to make sure it is not singular
        a += 4*eye(n)
        a_ = csc_array(a)
        lu = splu(a_)

        # And now test that we don't have a refcount bug
        rc = sys.getrefcount(lu)
        for attr in ('perm_r', 'perm_c'):
            perm = getattr(lu, attr)
            assert_equal(sys.getrefcount(lu), rc + 1)
            del perm
            assert_equal(sys.getrefcount(lu), rc)

    def test_bad_inputs(self):
        A = self.A.tocsc()
        rng = np.random.default_rng(235634)

        assert_raises(ValueError, splu, A[:,:4])
        assert_raises(ValueError, spilu, A[:,:4])

        for lu in [splu(A), spilu(A)]:
            b = rng.random(42)
            B = rng.random((42, 3))
            BB = rng.random((self.n, 3, 9))
            assert_raises(ValueError, lu.solve, b)
            assert_raises(ValueError, lu.solve, B)
            assert_raises(ValueError, lu.solve, BB)
            assert_raises(TypeError, lu.solve,
                          b.astype(np.complex64))
            assert_raises(TypeError, lu.solve,
                          b.astype(np.complex128))

    def test_superlu_dlamch_i386_nan(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", SparseEfficiencyWarning)
            # SuperLU 4.3 calls some functions returning floats without
            # declaring them. On i386@linux call convention, this fails to
            # clear floating point registers after call. As a result, NaN
            # can appear in the next floating point operation made.
            #
            # Here's a test case that triggered the issue.
            n = 8
            d = np.arange(n) + 1
            A = dia_array(((d, 2*d, d[::-1]), (-3, 0, 5)), shape=(n, n))
            A = A.astype(np.float32)
            spilu(A)
            A = A + 1j*A
            B = A.toarray()
            assert_(not np.isnan(B).any())

    def test_lu_attr(self):
        def check(dtype, complex_2=False):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", SparseEfficiencyWarning)

                A = self.A.astype(dtype)

                if complex_2:
                    A = A + 1j*A.T

                n = A.shape[0]
                lu = splu(A)

                # Check that the decomposition is as advertised

                Pc = np.zeros((n, n))
                Pc[np.arange(n), lu.perm_c] = 1

                Pr = np.zeros((n, n))
                Pr[lu.perm_r, np.arange(n)] = 1

                Ad = A.toarray()
                lhs = Pr.dot(Ad).dot(Pc)
                rhs = (lu.L @ lu.U).toarray()

                eps = np.finfo(dtype).eps

                assert_allclose(lhs, rhs, atol=100*eps)

        check(np.float32)
        check(np.float64)
        check(np.complex64)
        check(np.complex128)
        check(np.complex64, True)
        check(np.complex128, True)

    @pytest.mark.slow
    def test_threads_parallel(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", SparseEfficiencyWarning)
            oks = []

            def worker():
                try:
                    self.test_splu_basic()
                    self._internal_test_splu_smoketest()
                    self._internal_test_spilu_smoketest()
                    oks.append(True)
                except Exception:
                    pass

            threads = [threading.Thread(target=worker)
                       for k in range(20)]
            for t in threads:
                t.start()
            for t in threads:
                t.join()

            assert_equal(len(oks), 20)

    def test_singular_matrix(self):
        # Test that SuperLU does not print to stdout when a singular matrix is
        # passed. See gh-20993.
        A = eye_array(10, format='csr')
        A[-1, -1] = 0
        b = np.zeros(10)
        with pytest.warns(MatrixRankWarning):
            res = spsolve(A, b)
            assert np.isnan(res).all()


class TestGstrsErrors:
    def setup_method(self):
      self.A = csc_array([[1.0,2.0,3.0],[4.0,5.0,6.0],[7.0,8.0,9.0]], dtype=np.float64)
      self.b = np.array([[1.0],[2.0],[3.0]], dtype=np.float64)

    def test_trans(self):
        L = tril(self.A, format='csc')
        U = triu(self.A, k=1, format='csc')
        with assert_raises(ValueError, match="trans must be N, T, or H"):
            _superlu.gstrs('X', L.shape[0], L.nnz, L.data, L.indices, L.indptr,
                                U.shape[0], U.nnz, U.data, U.indices, U.indptr, self.b)

    def test_shape_LU(self):
        L = tril(self.A[0:2,0:2], format='csc')
        U = triu(self.A, k=1, format='csc')
        with assert_raises(ValueError, match="L and U must have the same dimension"):
            _superlu.gstrs('N', L.shape[0], L.nnz, L.data, L.indices, L.indptr,
                                U.shape[0], U.nnz, U.data, U.indices, U.indptr, self.b)

    def test_shape_b(self):
        L = tril(self.A, format='csc')
        U = triu(self.A, k=1, format='csc')
        with assert_raises(ValueError, match="right hand side array has invalid shape"):
            _superlu.gstrs('N', L.shape[0], L.nnz, L.data, L.indices, L.indptr,
                                U.shape[0], U.nnz, U.data, U.indices, U.indptr,
                                self.b[0:2])

    def test_types_differ(self):
        L = tril(self.A.astype(np.float32), format='csc')
        U = triu(self.A, k=1, format='csc')
        with assert_raises(TypeError, match="nzvals types of L and U differ"):
            _superlu.gstrs('N', L.shape[0], L.nnz, L.data, L.indices, L.indptr,
                                U.shape[0], U.nnz, U.data, U.indices, U.indptr, self.b)

    def test_types_unsupported(self):
        L = tril(self.A.astype(np.uint8), format='csc')
        U = triu(self.A.astype(np.uint8), k=1, format='csc')
        with assert_raises(TypeError, match="nzvals is not of a type supported"):
            _superlu.gstrs('N', L.shape[0], L.nnz, L.data, L.indices, L.indptr,
                                U.shape[0], U.nnz, U.data, U.indices, U.indptr,
                                self.b.astype(np.uint8))

class TestSpsolveTriangular:
    def setup_method(self):
        use_solver(useUmfpack=False)

    @pytest.mark.parametrize("fmt",["csr","csc"])
    def test_zero_diagonal(self,fmt):
        n = 5
        rng = np.random.default_rng(43876432987)
        A = coo_array(rng.standard_normal((n, n)))
        b = np.arange(n)
        A = tril(A, k=0, format=fmt)

        x = spsolve_triangular(A, b, unit_diagonal=True, lower=True)

        A.setdiag(1)
        assert_allclose(A.dot(x), b)

        # Regression test from gh-15199
        A = np.array([[0, 0, 0], [1, 0, 0], [1, 1, 0]], dtype=np.float64)
        b = np.array([1., 2., 3.])
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", "CSC or CSR matrix format is", SparseEfficiencyWarning)
            spsolve_triangular(A, b, unit_diagonal=True)

    @pytest.mark.parametrize("fmt",["csr","csc"])
    def test_singular(self,fmt):
        n = 5
        if fmt == "csr":
            A = csr_array((n, n))
        else:
            A = csc_array((n, n))
        b = np.arange(n)
        for lower in (True, False):
            assert_raises(scipy.linalg.LinAlgError,
                          spsolve_triangular, A, b, lower=lower)

    def test_bad_shape(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", SparseEfficiencyWarning)
            # A is not square.
            A = np.zeros((3, 4))
            b = ones((4, 1))
            assert_raises(ValueError, spsolve_triangular, A, b)
            # A2 and b2 have incompatible shapes.
            A2 = csr_array(eye(3))
            b2 = array([1.0, 2.0])
            assert_raises(ValueError, spsolve_triangular, A2, b2)

    def test_input_types(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", SparseEfficiencyWarning)
            A = array([[1., 0.], [1., 2.]])
            b = array([[2., 0.], [2., 2.]])
            for matrix_type in (array, csc_array, csr_array):
                x = spsolve_triangular(matrix_type(A), b, lower=True)
                assert_array_almost_equal(A.dot(x), b)

    @pytest.mark.slow
    @pytest.mark.parametrize("n", [10, 10**2, 10**3])
    @pytest.mark.parametrize("m", [1, 10])
    @pytest.mark.parametrize("lower", [True, False])
    @pytest.mark.parametrize("format", ["csr", "csc"])
    @pytest.mark.parametrize("unit_diagonal", [False, True])
    @pytest.mark.parametrize("choice_of_A", ["real", "complex"])
    @pytest.mark.parametrize("choice_of_b", ["floats", "ints", "complexints"])
    def test_random(self, n, m, lower, format, unit_diagonal, choice_of_A, choice_of_b):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", SparseEfficiencyWarning)
            def random_triangle_matrix(n, lower=True, format="csr", choice_of_A="real"):
                if choice_of_A == "real":
                    dtype = np.float64
                elif choice_of_A == "complex":
                    dtype = np.complex128
                else:
                    raise ValueError("choice_of_A must be 'real' or 'complex'.")
                rng = np.random.default_rng(789002319)
                rvs = rng.random
                A = random_array((n, n), density=0.1, format='lil', dtype=dtype,
                                 random_state=rng, data_sampler=rvs)
                if lower:
                    A = tril(A, format="lil")
                else:
                    A = triu(A, format="lil")
                for i in range(n):
                    A[i, i] = np.random.rand() + 1
                if format == "csc":
                    A = A.tocsc(copy=False)
                else:
                    A = A.tocsr(copy=False)
                return A

            rng = np.random.default_rng(1234)
            A = random_triangle_matrix(n, lower=lower)
            if choice_of_b == "floats":
                b = rng.random((n, m))
            elif choice_of_b == "ints":
                b = rng.integers(-9, 9, (n, m))
            elif choice_of_b == "complexints":
                b = rng.integers(-9, 9, (n, m)) + rng.integers(-9, 9, (n, m)) * 1j
            else:
                raise ValueError(
                    "choice_of_b must be 'floats', 'ints', or 'complexints'.")
            x = spsolve_triangular(A, b, lower=lower, unit_diagonal=unit_diagonal)
            if unit_diagonal:
                A.setdiag(1)
            assert_allclose(A.dot(x), b, atol=1.5e-6)


@pytest.mark.parametrize("nnz", [10, 10**2, 10**3])
@pytest.mark.parametrize("fmt", ["csr", "csc", "coo", "dia", "dok", "lil"])
def test_is_sptriangular_and_spbandwidth(nnz, fmt):
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', SparseEfficiencyWarning)
        rng = np.random.default_rng(42)

        N = nnz // 2
        dens = 0.1
        A = random_array((N, N), density=dens, format="csr", rng=rng)
        A[1, 3] = A[3, 1] = 22  # ensure not upper or lower
        A = A.asformat(fmt)
        AU = triu(A, format=fmt)
        AL = tril(A, format=fmt)
        D = 0.1 * eye_array(N, format=fmt)

        assert is_sptriangular(A) == (False, False)
        assert is_sptriangular(AL) == (True, False)
        assert is_sptriangular(AU) == (False, True)
        assert is_sptriangular(D) == (True, True)

        assert spbandwidth(A) == scipy.linalg.bandwidth(A.toarray())
        assert spbandwidth(AU) == scipy.linalg.bandwidth(AU.toarray())
        assert spbandwidth(AL) == scipy.linalg.bandwidth(AL.toarray())
        assert spbandwidth(D) == scipy.linalg.bandwidth(D.toarray())



"""
Provides classes to style the axis lines.
"""
import math

import numpy as np

import matplotlib as mpl
from matplotlib.patches import _Style, FancyArrowPatch
from matplotlib.path import Path
from matplotlib.transforms import IdentityTransform


class _FancyAxislineStyle:
    class SimpleArrow(FancyArrowPatch):
        """The artist class that will be returned for SimpleArrow style."""
        _ARROW_STYLE = "->"

        def __init__(self, axis_artist, line_path, transform,
                     line_mutation_scale):
            self._axis_artist = axis_artist
            self._line_transform = transform
            self._line_path = line_path
            self._line_mutation_scale = line_mutation_scale

            FancyArrowPatch.__init__(self,
                                     path=self._line_path,
                                     arrowstyle=self._ARROW_STYLE,
                                     patchA=None,
                                     patchB=None,
                                     shrinkA=0.,
                                     shrinkB=0.,
                                     mutation_scale=line_mutation_scale,
                                     mutation_aspect=None,
                                     transform=IdentityTransform(),
                                     )

        def set_line_mutation_scale(self, scale):
            self.set_mutation_scale(scale*self._line_mutation_scale)

        def _extend_path(self, path, mutation_size=10):
            """
            Extend the path to make a room for drawing arrow.
            """
            (x0, y0), (x1, y1) = path.vertices[-2:]
            theta = math.atan2(y1 - y0, x1 - x0)
            x2 = x1 + math.cos(theta) * mutation_size
            y2 = y1 + math.sin(theta) * mutation_size
            if path.codes is None:
                return Path(np.concatenate([path.vertices, [[x2, y2]]]))
            else:
                return Path(np.concatenate([path.vertices, [[x2, y2]]]),
                            np.concatenate([path.codes, [Path.LINETO]]))

        def set_path(self, path):
            self._line_path = path

        def draw(self, renderer):
            """
            Draw the axis line.
             1) Transform the path to the display coordinate.
             2) Extend the path to make a room for arrow.
             3) Update the path of the FancyArrowPatch.
             4) Draw.
            """
            path_in_disp = self._line_transform.transform_path(self._line_path)
            mutation_size = self.get_mutation_scale()  # line_mutation_scale()
            extended_path = self._extend_path(path_in_disp,
                                              mutation_size=mutation_size)
            self._path_original = extended_path
            FancyArrowPatch.draw(self, renderer)

        def get_window_extent(self, renderer=None):

            path_in_disp = self._line_transform.transform_path(self._line_path)
            mutation_size = self.get_mutation_scale()  # line_mutation_scale()
            extended_path = self._extend_path(path_in_disp,
                                              mutation_size=mutation_size)
            self._path_original = extended_path
            return FancyArrowPatch.get_window_extent(self, renderer)

    class FilledArrow(SimpleArrow):
        """The artist class that will be returned for FilledArrow style."""
        _ARROW_STYLE = "-|>"

        def __init__(self, axis_artist, line_path, transform,
                     line_mutation_scale, facecolor):
            super().__init__(axis_artist, line_path, transform,
                             line_mutation_scale)
            self.set_facecolor(facecolor)


class AxislineStyle(_Style):
    """
    A container class which defines style classes for AxisArtists.

    An instance of any axisline style class is a callable object,
    whose call signature is ::

       __call__(self, axis_artist, path, transform)

    When called, this should return an `.Artist` with the following methods::

      def set_path(self, path):
          # set the path for axisline.

      def set_line_mutation_scale(self, scale):
          # set the scale

      def draw(self, renderer):
          # draw
    """

    _style_list = {}

    class _Base:
        # The derived classes are required to be able to be initialized
        # w/o arguments, i.e., all its argument (except self) must have
        # the default values.

        def __init__(self):
            """
            initialization.
            """
            super().__init__()

        def __call__(self, axis_artist, transform):
            """
            Given the AxisArtist instance, and transform for the path (set_path
            method), return the Matplotlib artist for drawing the axis line.
            """
            return self.new_line(axis_artist, transform)

    class SimpleArrow(_Base):
        """
        A simple arrow.
        """

        ArrowAxisClass = _FancyAxislineStyle.SimpleArrow

        def __init__(self, size=1):
            """
            Parameters
            ----------
            size : float
                Size of the arrow as a fraction of the ticklabel size.
            """

            self.size = size
            super().__init__()

        def new_line(self, axis_artist, transform):

            linepath = Path([(0, 0), (0, 1)])
            axisline = self.ArrowAxisClass(axis_artist, linepath, transform,
                                           line_mutation_scale=self.size)
            return axisline

    _style_list["->"] = SimpleArrow

    class FilledArrow(SimpleArrow):
        """
        An arrow with a filled head.
        """

        ArrowAxisClass = _FancyAxislineStyle.FilledArrow

        def __init__(self, size=1, facecolor=None):
            """
            Parameters
            ----------
            size : float
                Size of the arrow as a fraction of the ticklabel size.
            facecolor : :mpltype:`color`, default: :rc:`axes.edgecolor`
                Fill color.

                .. versionadded:: 3.7
            """

            facecolor = mpl._val_or_rc(facecolor, 'axes.edgecolor')
            self.size = size
            self._facecolor = facecolor
            super().__init__(size=size)

        def new_line(self, axis_artist, transform):
            linepath = Path([(0, 0), (0, 1)])
            axisline = self.ArrowAxisClass(axis_artist, linepath, transform,
                                           line_mutation_scale=self.size,
                                           facecolor=self._facecolor)
            return axisline

    _style_list["-|>"] = FilledArrow
